# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.structure.io"
__author__ = "Patrick Kunzmann"
__all__ = ["TrajectoryFile"]

import itertools
import abc
import numpy as np
from ..atoms import AtomArray, AtomArrayStack, stack, from_template
from ...file import File


class TrajectoryFile(File, metaclass=abc.ABCMeta):
    """
    This file class represents a trajectory file interfacing a
    trajectory file class from `MDtraj`.
    
    A trajectory file stores atom coordinates over multiple (time)
    frames. The file formats are usually binary and involve sometimes
    heavy compression, so that a large number of frames can be stored
    in relatively small space.
    Since all :class:`TrajectoryFile` subclasses interface *MDtraj*
    trajectory file classes, `MDtraj` must be installed to use any of
    them.

    Notes
    -----
    When extracting data from the file, only a reference to the
    data arrays stored in this file are created.
    The same is true, when setting data in the file.
    Therefore, it is strongly recommended to make a copy of the
    respective array, if the array is modified.
    """
    
    def __init__(self):
        super().__init__()
        self._coord = None
        self._time = None
        self._box = None
        self._model_count = None
    

    @classmethod
    def read(cls, file_name, start=None, stop=None, step=None,
             atom_i=None, chunk_size=None):
        """
        Read a trajectory file.
        
        A trajectory file can be seen as a file representation of an
        :class:`AtomArrayStack`.
        Therefore, `start`, `stop` and `step` represent slice parameters
        of the index of the first dimension and
        `atom_i` represents an index array for the second dimension.
        
        Parameters
        ----------
        file_name : str
            The path of the file to be read.
            A file-like-object cannot be used.
        start : int, optional
            The frame index, where file parsing is started. If no value
            is given, parsing starts at the first frame.
            The index starts at 0.
        stop : int, optional
            The exclusive frame index, where file parsing ends.
            If no value is given, parsing stops after the last frame.
            The index starts at 0.
        step : int, optional
            If this value is set, the function reads only every n-th
            frame from the file.
        atom_i : ndarray, dtype=int, optional
            If this parameter is set, only the atoms at the given
            indices are read from each frame.
        chunk_size : int, optional
            If this parameter is set, the trajectory is read in chunks:
            Only the number of frames specified by this parameter are
            read at once.
            The resulting chunks of frames are automatically
            concatenated, after all chunks are collected.
            Use this parameter, if a :class:`MemoryError` is raised
            when a trajectory file is read.
            Although lower values can decrease the memory consumption of
            reading trajectories, they also increase the computation
            time.
        
        Returns
        -------
        file_object : TrajectoryFile
            The parsed trajectory file.
        """
        file = cls()

        if chunk_size is not None:
            if chunk_size < 1:
                raise ValueError("Chunk size must be greater than 0")
            # Chunk size must be a multiple of step size to ensure that
            # the step distance at the chunk border is the same as
            # within a chunk
            # -> round chunk size up to a multiple of step size
            if step is not None and chunk_size % step != 0:
                chunk_size = ((chunk_size // step) + 1) * step

        traj_type = cls.traj_type()
        with traj_type(file_name, "r") as f:
            
            if start is None:
                start = 0
            # Discard atoms before start
            if start != 0:
                if chunk_size is None or chunk_size > start:
                    f.read(n_frames=start, stride=None, atom_indices=atom_i)
                else:
                    TrajectoryFile._read_chunk_wise(
                        f, start, None, atom_i, chunk_size, discard=True
                    )
            
            # The upcoming frames are saved
            # Calculate the amount of frames to be read
            if stop is None:
                n_frames = None
            else:
                n_frames = stop-start
            if step is not None and n_frames is not None:
                # Divide number of frames by 'step' in order to convert
                # 'step' into 'stride'
                # Since the 0th frame is always included,
                # the number of frames is decremented before division
                # and incremented afterwards again
                n_frames = ((n_frames - 1) // step) + 1
            
            # Read frames
            if chunk_size is None:
                result = f.read(n_frames, stride=step, atom_indices=atom_i)
            else:
                result = TrajectoryFile._read_chunk_wise(
                    f, n_frames, step, atom_i, chunk_size, discard=False
                )
        
        # nm to Angstrom
        coord, box, time = cls.process_read_values(result)
        file.set_coord(coord)
        file.set_box(box)
        file.set_time(time)

        return file
    

    @classmethod
    def read_iter(cls, file_name, start=None, stop=None, step=None,
                  atom_i=None):
        """
        Create an iterator over each frame of the given trajectory file
        in the selected range.
        
        Parameters
        ----------
        file_name : str
            The path of the file to be read.
            A file-like-object cannot be used.
        start : int, optional
            The frame index, where file parsing is started. If no value
            is given, parsing starts at the first frame.
            The index starts at 0.
        stop : int, optional
            The exclusive frame index, where file parsing ends.
            If no value is given, parsing stops at the end of file.
            The index starts at 0.
        step : int, optional
            If this value is set, the function reads only every n-th
            frame from the file.
        atom_i : ndarray, dtype=int, optional
            If this parameter is set, only the atoms at the given
            indices are read from each frame.
        
        Yields
        ------
        coord : ndarray, dtype=float32, shape=(n,3)
            The atom coordinates in the current frame.
        box : ndarray, dtype=float32, shape=(3,3)
            The box vectors of the current frame.
        time : float
            the simlation time of the current frame in *ps*.
        
        See also
        --------
        read_iter_structure
        
        Notes
        -----
        The `step` parameter does currently not work for *DCD* files.
        """
        traj_type = cls.traj_type()
        with traj_type(file_name, "r") as f:
            
            if start is None:
                start = 0
            # Discard atoms before start
            if start != 0:
                f.read(n_frames=start, stride=None, atom_indices=atom_i)
            
            # The upcoming frames are read
            # Calculate the amount of frames to be read
            if stop is None:
                n_frames = None
            else:
                n_frames = stop-start
            if step is not None and n_frames is not None:
                # Divide number of frames by 'step' in order to convert
                # 'step' into 'stride'
                # Since the 0th frame is always included,
                # the number of frames is decremented before division
                # and incremented afterwards again
                n_frames = ((n_frames - 1) // step) + 1
            
            # Read frames
            frame_i = 0
            while True:
                if n_frames is not None and frame_i >= n_frames:
                    # Stop frame reached -> stop interation
                    break
                # Read one frame per 'yield'
                result = f.read(1, stride=step, atom_indices=atom_i)
                if len(result[0]) == 0:
                    # Empty array was read
                    # -> no frames left -> stop interation
                    break
                coord, box, time = cls.process_read_values(result)
                # Only one frame
                # -> only one element in first dimension
                # -> remove first dimension
                coord = coord[0]
                box = box[0] if box is not None else None
                time = float(time[0]) if time is not None else None
                yield coord, box, time
                frame_i += 1
    

    @classmethod
    def read_iter_structure(cls, file_name, template, start=None, stop=None,
                            step=None, atom_i=None):
        """
        Create an iterator over each frame of the given trajectory file
        in the selected range.

        In contrast to :func:`read_iter()`, this function creates an
        iterator over the structure as :class:`AtomArray`.
        Since trajectory files usually only contain atom coordinate
        information and no topology information, this method requires
        a template atom array or stack. This template can be acquired
        for example from a PDB file, which is associated with the
        trajectory file. 
        
        Parameters
        ----------
        file_name : str
            The path of the file to be read.
            A file-like-object cannot be used.
        template : AtomArray or AtomArrayStack
            The template array or stack, where the atom annotation data
            is taken from.
        start : int, optional
            The frame index, where file parsing is started. If no value
            is given, parsing starts at the first frame.
            The index starts at 0.
        stop : int, optional
            The exclusive frame index, where file parsing ends.
            If no value is given, parsing stops at the end of file.
            The index starts at 0.
        step : int, optional
            If this value is set, the function reads only every n-th
            frame from the file.
        atom_i : ndarray, dtype=int, optional
            If this parameter is set, only the atoms at the given
            indices are read from each frame.
        
        Yields
        ------
        structure : AtomArray
            The structure of the current frame.
        
        See also
        --------
        read_iter
        
        Notes
        -----
        This iterator creates a new copy of the given template for every
        frame.
        If a higher efficiency is required, please use the
        :func:`read_iter()` function.

        The `step` parameter does currently not work for *DCD* files.
        """
        if isinstance(template, AtomArrayStack):
            template = template[0]
        elif not isinstance(template, AtomArray):
            raise TypeError(
                f"An 'AtomArray' or 'AtomArrayStack' is expected as template, "
                f"not '{type(template).__name__}'"
            )
        for coord, box, _ in cls.read_iter(
            file_name, start, stop, step, atom_i
        ):
            frame = template.copy()
            frame.coord = coord
            frame.box = box
            yield frame

    
    def write(self, file_name):
        """
        Write the content into a trajectory file.

        Parameters
        ----------
        file_name : str
            The path of the file to be written to.
            A file-like-object cannot be used.
        """
        traj_type = self.traj_type()
        param = self.prepare_write_values(self._coord, self._box, self._time)
        with traj_type(file_name, 'w') as f:
            f.write(**param)
    

    @classmethod
    def write_iter(cls, file_name, coord, box=None, time=None):
        """
        Iterate over the given `coord` and write each item into
        the file specified by `file_name`.

        In contrast to :meth:`write()`, the data is not stored in an
        intermediate :class:`TrajectoryFile`, but is directly written
        to the file.
        Hence, this class method may save a large amount of memory if
        a large file should be written, if `coord` are provided as
        generator.
        
        Parameters
        ----------
        file_name : str
            The path of the file to be written to.
            A file-like-object cannot be used.
        coord : generator or array-like of ndarray, shape=(n,3), dtype=float
            The atom coordinates for each frame.
        box : generator or array-like of ndarray, shape=(3,3), dtype=float, optional
            The three box vectors for each frame.
        time : generator or array-like of float, optional
            The simulation time in *ps* for each frame.

        Notes
        -----
        The `time` parameter has no effect for *TNG* and *DCD* files.
        """
        if box is None:
            box = itertools.repeat(None)
        if time is None:
            time = itertools.repeat(None)

        traj_type = cls.traj_type()
        with traj_type(file_name, 'w') as f:
            for c, b, t in zip(coord, box, time):
                if c.ndim != 2:
                    raise IndexError(
                        f"Expected ndarray with 2 dimensions, got {c.ndim}"
                    )
                # Add new dimension of length one
                # for compatibility with 'prepare_write_values()'
                c = c[np.newaxis, :]
                if b is not None:
                    b = np.expand_dims(b, axis=0)
                if t is not None:
                    t = np.expand_dims(t, axis=0)
                param = cls.prepare_write_values(c, b, t)
                f.write(**param)
    

    def get_coord(self):
        """
        Extract only the atom coordinates from the trajectory file.
        
        Returns
        -------
        coord : ndarray, dtype=float, shape=(m,n,3)
            The coordinates stored in the trajectory file.
        """
        return self._coord
    

    def get_time(self):
        """
        Get the simlation time in *ps* values for each frame.
        
        Returns
        -------
        time : ndarray, dtype=float, shape=(m,)
            A one dimensional array containing the time values for the
            frames, that were read from the file.
        """
        return self._time
    

    def get_box(self):
        """
        Get the box vectors for each frame.
        
        Returns
        -------
        box : ndarray, dtype=float, shape=(m,3,3)
            An array containing the box dimensions for the
            frames, that were read from the file.
        """
        return self._box
    

    def set_coord(self, coord):
        """
        Set the atom coordinates in the trajectory file.
        
        Parameters
        ----------
        coord : ndarray, dtype=float, shape=(m,n,3)
            The coordinates to be set.
        """
        self._check_model_count(coord)
        self._coord = coord
    

    def set_time(self, time):
        """
        Set the simulation time of each frame in the trajectory file.
        
        Parameters
        ----------
        time : ndarray, dtype=float, shape=(m,)
            The simulation time in *ps* to be set.
        """
        self._check_model_count(time)
        self._time = time
    

    def set_box(self, box):
        """
        Set the periodic box vectors of each frame in the trajectory
        file.
        
        Parameters
        ----------
        time : ndarray, dtype=float, shape=(m,3,3)
            The box vectors to be set.
        """
        self._check_model_count(box)
        self._box = box
    

    def get_structure(self, template):
        """
        Convert the trajectory file content into an
        :class:`AtomArrayStack`.
        
        Since trajectory files usually only contain atom coordinate
        information and no topology information, this method requires
        a template atom array or stack. This template can be acquired
        for example from a PDB file, which is associated with the
        trajectory file. 
        
        Parameters
        ----------
        template : AtomArray or AtomArrayStack
            The template array or stack, where the atom annotation data
            is taken from.
        
        Returns
        -------
        array_stack : AtomArrayStack
            A stack containing the annontation arrays from `template`
            but the coordinates and the simulation boxes from the
            trajectory file.
        """
        return from_template(template, self.get_coord(), self.get_box())
    

    def set_structure(self, structure, time=None):
        """
        Write an atom array (stack) into the trajectory file object.
        
        The topology information (chain, residue, etc.) is not saved in
        the file.
        
        Parameters
        ----------
        structure : AtomArray or AtomArrayStack
            The structure to be put into the trajectory file.
        time : ndarray, dtype=float, shape=(n,), optional
            The simulation time for each frame in `structure`.
        """
        coord = structure.coord
        box = structure.box
        if coord.ndim == 2:
            coord = coord[np.newaxis, :, :]
        if box is not None and box.ndim == 2:
            box = box[np.newaxis, :, :]
        self.set_coord(coord)
        if box is not None:
            self.set_box(box)
        if time is not None:
            self.set_time(time)

    
    def copy(self):
        """
        This operation is not implemented for trajectory files.
        
        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError("Copying is not implemented "
                                  "for trajectory files")
    

    @classmethod
    @abc.abstractmethod
    def traj_type(cls):
        """
        The `MDtraj` files class to be used.
        
        PROTECTED: Override when inheriting.
        
        Returns
        -------
        class
            An `MDtraj` subclass of :class:`TrajectoryFile`.
        """
        pass
    

    @classmethod
    @abc.abstractmethod
    def process_read_values(cls, read_values):
        """
        Convert the return value of the `read()` method of the
        respective :class:`mdtraj.TrajectoryFile` into coordinates,
        simulation box and simulation time.
        
        PROTECTED: Override when inheriting.
        
        Parameters
        ----------
        read_values : tuple
            The return value of the respective
            :func:`mdtraj.TrajectoryFile.read()` method.
        
        Returns
        -------
        coord : ndarray, dtype=float, shape=(m,n,3)
            The atom coordinates in Å for each frame.
        box : ndarray, dtype=float, shape=(m,3,3) or None
            The box vectors in Å for each frame.
        time : ndarray, dtype=float, shape=(m,) or None
            The simulation time in ps for each frame.
        """
        pass
    

    @classmethod
    @abc.abstractmethod
    def prepare_write_values(cls, coord, box, time):
        """
        Convert the `coord`, `box` and `time` attribute into a
        dictionary that is given as *kwargs* to the respective
        :func:`mdtraj.TrajectoryFile.write()` method.

        PROTECTED: Override when inheriting.

        Parameters
        ----------
        coord : ndarray, dtype=float, shape=(m,n,3)
            The atom coordinates in Å for each frame.
        box : ndarray, dtype=float, shape=(m,3,3)
            The box vectors in Å for each frame.
        time : ndarray, dtype=float, shape=(m,)
            The simulation time in ps for each frame.
        
        Returns
        -------
        parameters : dict
            This dictionary is given as *kwargs* parameter to the
            respective :func:`mdtraj.TrajectoryFile.write()` method.
        """
        pass


    def _check_model_count(self, array):
        """
        Check if the amount of models in the given array is equal to
        the amount of models in the file.
        If not, raise an exception.
        If the amount of models in the file is not set yet, set it with
        the amount of models in the array.
        """
        if array is None:
            return
        if self._model_count is None:
            self._model_count = len(array)
        else:
            if self._model_count != len(array):
                raise ValueError(
                    f"{len(array)} models were given, "
                    f"but the file contains {self._model_count} models"
                )
    

    @staticmethod
    def _read_chunk_wise(file, n_frames, step, atom_i, chunk_size,
                         discard=False):
        """
        Similar to :func:`read()`, just for chunk-wise reading of the
        trajectory.

        `n_frames` is already the actual number of frames in the outout
        arrays, i.e. the original number was divided by `step`.
        """
        chunks = []
        remaining_frames = n_frames
        # If n_frames is None, this is condition is never False
        # -> break out of loop when read chunk is empty (see below)
        while remaining_frames != 0:
            if remaining_frames is not None:
                n = min(remaining_frames, chunk_size)
            else:
                n = chunk_size
            try:
                chunk = file.read(n_frames=n, stride=step, atom_indices=atom_i)
            except ValueError as e:
                # MDTraj raises exception because no coordinates can be
                # concatenated
                # -> all frames have been read
                # -> stop reading chunks
                if str(e) != "need at least one array to concatenate":
                    raise
                else:
                    break
            if len(chunk[0]) == 0:
                # Coordinates have a length of 0
                # -> all frames have been read
                # -> stop reading chunks
                break
            if not discard:
                chunks.append(chunk)
            if remaining_frames is not None:
                remaining_frames -= n
        
        if not discard:
            # Assemble the chunks into contiguous arrays
            # for each value (coord, box, time)
            result = [None] * len(chunks[0])
            # Iterate over all valuesin the result tuple
            # and concatenate the corresponding value from each chunk,
            # if the value is not None
            # The amount of values is determined from the first chunk
            for i in range(len(chunks[0])):
                if chunks[0][i] is not None:
                    result[i] = np.concatenate([chunk[i] for chunk in chunks])
                else:
                    result[i] = None
            return tuple(result)
        else:
            return None


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import glob
import itertools
from os.path import join
import numpy as np
import pytest
import biotite.structure as struc
import biotite.structure.io as strucio
import biotite.database.rcsb as rcsb
import biotite.structure as struc
from ..util import data_dir


@pytest.mark.parametrize(
    "path", glob.glob(join(data_dir("structure"), "*.mmtf"))
)
def test_superimposition_array(path):
    """
    Take a structure and rotate and translate a copy of it, so that they
    are not superimposed anymore.
    Then superimpose these structure onto each other and expect an
    almost perfect match.
    """
    fixed = strucio.load_structure(path, model=1)
    
    mobile = fixed.copy()
    mobile = struc.rotate(mobile, (1,2,3))
    mobile = struc.translate(mobile, (1,2,3))
    
    fitted, transformation = struc.superimpose(
        fixed, mobile
    )
    
    assert struc.rmsd(fixed, fitted) == pytest.approx(0, abs=6e-4)
    
    fitted = struc.superimpose_apply(mobile, transformation)
    
    assert struc.rmsd(fixed, fitted) == pytest.approx(0, abs=6e-4)


@pytest.mark.parametrize("ca_only", (True, False))
def test_superimposition_stack(ca_only):
    """
    Take a structure with multiple models where each model is not
    (optimally) superimposed onto each other.
    Then superimpose and expect an improved RMSD.
    """
    path = join(data_dir("structure"), "1l2y.mmtf")
    stack = strucio.load_structure(path)
    fixed = stack[0]
    mobile = stack[1:]
    if ca_only:
        mask = (mobile.atom_name == "CA")
    else:
        mask = None
    
    fitted, transformation = struc.superimpose(fixed, mobile, mask)
    
    if ca_only:
        # The superimpositions are better for most cases than the
        # superimpositions in the structure file
        # -> Use average
        assert np.mean(struc.rmsd(fixed, fitted)) \
             < np.mean(struc.rmsd(fixed, mobile))
    else:
        # The superimpositions are better than the superimpositions
        # in the structure file
        assert (struc.rmsd(fixed, fitted) < struc.rmsd(fixed, mobile)).all()


@pytest.mark.parametrize("seed", range(5))
def test_masked_superimposition(seed):
    """
    Take two models of the same structure and superimpose based on a
    single, randomly chosen atom.
    Since two atoms can be superimposed perfectly, the distance between
    the atom in both models should be 0.
    """

    path = join(data_dir("structure"), "1l2y.mmtf")
    fixed = strucio.load_structure(path, model=1)
    mobile = strucio.load_structure(path, model=2)

    # Create random mask for a single atom
    np.random.seed(seed)
    mask = np.full(fixed.array_length(), False)
    mask[np.random.randint(fixed.array_length())] = True
    
    # The distance between the atom in both models should not be
    # already 0 prior to superimposition
    assert struc.distance(fixed[mask], mobile[mask])[0] \
        != pytest.approx(0, abs=5e-4)

    fitted, transformation = struc.superimpose(
        fixed, mobile, mask
    )
    
    assert struc.distance(fixed[mask], fitted[mask])[0] \
        == pytest.approx(0, abs=5e-4)
    
    fitted = struc.superimpose_apply(mobile, transformation)
    
    struc.distance(fixed[mask], fitted[mask])[0] \
        == pytest.approx(0, abs=5e-4)


def test_ndarray_inputs_for_superimpose_raise_value_error():
    """
    Check that using incorrect types as inputs fails.
    """
    fixed = np.random.rand(3, 3)
    mobile = np.random.rand(3, 3)
    with pytest.raises(ValueError):
        struc.superimpose(fixed, mobile)


@pytest.mark.parametrize(
    "single_model, single_atom", itertools.product([False, True], [False, True])
)
def test_input_shapes(single_model, single_atom):
    """
    Test whether :func:`superimpose()` infers the correct output shape,
    even if the input :class:`AtomArrayStack` contains only a single
    model or a single atom.
    """
    path = join(data_dir("structure"), "1l2y.mmtf")
    stack = strucio.load_structure(path)
    fixed = stack[0]
    
    mobile = stack
    if single_model:
        mobile = mobile[:1, :]
    if single_atom:
        mobile = mobile[:, :1]
        fixed = fixed[:1]
    
    fitted, _ = struc.superimpose(fixed, mobile)

    assert type(fitted) == type(mobile)
    assert fitted.coord.shape == mobile.coord.shape



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import string
import pickle
import numpy as np
import pytest
import biotite.sequence as seq
import biotite.sequence.align as align


@pytest.fixture
def k():
    return 8

@pytest.fixture
def alphabet():
    return seq.NucleotideSequence.unambiguous_alphabet()

@pytest.fixture
def random_sequences(k, alphabet):
    N_SEQS = 10
    SEQ_LENGTH = 1000

    np.random.seed(0)
    sequences = []
    for _ in range(N_SEQS):
        sequence = seq.NucleotideSequence()
        sequence.code = np.random.randint(len(alphabet), size=SEQ_LENGTH)
        sequences.append(sequence)
    return sequences



@pytest.mark.parametrize("spacing", [None, "10111011011", "1001111010101"])
def test_from_sequences(k, random_sequences, spacing):
    """
    Test the :meth:`from_sequences()` constructor, by checking for each
    sequence position, if the position is in the C-array of the
    corresponding k-mer.
    """
    table = align.KmerTable.from_sequences(
        k, random_sequences, spacing=spacing
    )
    kmer_alph = align.KmerAlphabet(random_sequences[0].alphabet, k, spacing)
    assert kmer_alph == table.kmer_alphabet

    for i, sequence in enumerate(random_sequences):
        for j in range(kmer_alph.kmer_array_length(len(sequence))):
            if spacing is None:
                kmer = kmer_alph.fuse(sequence.code[j : j+k])
            else:
                kmer = kmer_alph.fuse(sequence.code[kmer_alph.spacing + j])
            assert np.array([i,j]) in table[kmer]



def test_from_kmers(k, random_sequences):
    """
    Test the :meth:`from_kmers()` constructor, by comparing it with an
    equivalent table created with :meth:`from_sequences()`.
    """
    ref_table = align.KmerTable.from_sequences(k, random_sequences)
    kmer_alph = ref_table.kmer_alphabet

    kmer_arrays = [
        kmer_alph.create_kmers(sequence.code) for sequence in random_sequences
    ]
    test_table = align.KmerTable.from_kmers(kmer_alph, kmer_arrays)
    
    assert test_table == ref_table


def test_from_tables(k, random_sequences):
    """
    Test :meth:`from_tables()` constructor by comparing the merge of
    single tables with adding sequences directly into a single table.
    """
    
    ref_table = align.KmerTable.from_sequences(
        k, random_sequences, np.arange(len(random_sequences))
    )

    tables = [
        align.KmerTable.from_sequences(k, [sequence], [i])
        for i, sequence in enumerate(random_sequences)
    ]
    test_table = align.KmerTable.from_tables(tables)
    
    assert test_table == ref_table


def test_from_positions(k, random_sequences):
    """
    Test :meth:`from_positions()` constructor by comparing it with an
    equivalent table created with :meth:`from_sequences()`.
    """
    ref_table = align.KmerTable.from_sequences(k, random_sequences)

    kmer_dict = {kmer : ref_table[kmer] for kmer in range(len(ref_table))}
    test_table = align.KmerTable.from_positions(
        ref_table.kmer_alphabet, kmer_dict
    )
    
    assert test_table == ref_table


@pytest.mark.parametrize("use_similarity_rule", [False, True])
def test_match_table(use_similarity_rule):
    """
    Test the :meth:`match_table()` method based on a known example.
    
    Using the similarity rule should give the same result, as it is
    chosen to yield only the same k-mer as similar k-mer.
    """
    alphabet = seq.LetterAlphabet(string.ascii_lowercase + "_")
    phrase1 = "how_much_wood_would_a_woodchuck_chuck_if_a_woodchuck_could_" \
              "chuck_wood"
    phrase2 = "woodchuck"
    sequence1 = seq.GeneralSequence(alphabet, phrase1)
    sequence2 = seq.GeneralSequence(alphabet, phrase2)

    rule = _identity_rule(alphabet) if use_similarity_rule else None
    
    table1 = align.KmerTable.from_sequences(4, [sequence1])
    table2 = align.KmerTable.from_sequences(4, [sequence2])

    ref_matches = set([
        (0,  9),
        (0, 22),
        (1, 23),
        (2, 24),
        (3, 25),
        (4, 26),
        (5, 27),
        (4, 32),
        (5, 33),
        (0, 43),
        (1, 44),
        (2, 45),
        (3, 46),
        (4, 47),
        (5, 48),
        (4, 59),
        (5, 60),
        (0, 65),
    ])

    test_matches = table1.match_table(table2, similarity_rule=rule)
    # the reference indices are irrelevant for this test
    test_matches = test_matches[:, [1,3]]
    test_matches = set([tuple(match) for match in test_matches])
    assert test_matches == ref_matches


@pytest.mark.parametrize("use_similarity_rule", [False, True])
def test_match(k, random_sequences, use_similarity_rule):
    """
    Test the :meth:`match()` method compared to a manual,
    inefficient approach using indexing.

    Using the similarity rule should give the same result, as it is
    chosen to yield only the same k-mer as similar k-mer.
    """
    query_sequence = random_sequences[0]
    table_sequences = random_sequences[1:]   
    table = align.KmerTable.from_sequences(k, table_sequences)
    
    kmers = table.kmer_alphabet.create_kmers(query_sequence.code)
    ref_matches = []
    for i, kmer in enumerate(kmers):
        matches = table[kmer]
        matches = np.stack(
            [
                np.full(len(matches), i, dtype=np.uint32), 
                matches[:,0],
                matches[:,1]
            ],
            axis=1
        )
        ref_matches.append(matches)
    ref_matches = np.concatenate(ref_matches)

    rule = _identity_rule(table.alphabet) if use_similarity_rule else None
    test_matches = table.match(query_sequence, similarity_rule=rule)

    assert np.array_equal(test_matches.tolist(), ref_matches.tolist())


@pytest.mark.parametrize("use_mask", [False, True])
def test_match_equivalence(k, random_sequences, use_mask):
    """
    Check if both, the :meth:`match()` and meth:`match_table()`
    method, find the same matches.
    """
    query_sequence = random_sequences[0]
    table_sequences = random_sequences[1:]

    if use_mask:
        # Create a random removal mask with low removal probability
        np.random.seed(0)
        removal_masks = [
            np.random.choice([False, True], size=len(sequence), p=[0.95, 0.05])
            for sequence in random_sequences
        ]
    else:
        removal_masks = [None] * len(random_sequences)
    query_mask = removal_masks[0]
    table_masks = removal_masks[1:]
    
    table = align.KmerTable.from_sequences(
        k, table_sequences, ignore_masks=table_masks
    )
    
    # 42 -> Dummy value that is distinct from all reference indices
    ref_table = align.KmerTable.from_sequences(
        k, [query_sequence], [42], ignore_masks=[query_mask]
    )
    ref_matches = table.match_table(ref_table)
    assert np.all(ref_matches[:,0] == 42)
    # Store matches in set to remove the order dependency
    # The first column is not present in the matches
    # returned by 'match_sequence()' -> [:, 1:]
    ref_matches = set([tuple(match) for match in ref_matches[:, 1:]])

    test_matches = table.match(
        query_sequence, ignore_mask=query_mask
    )
    test_matches = set([tuple(match) for match in test_matches])


    # Check if any match is found at all
    assert len(ref_matches) > 0
    # The first column is not present in 'test_matches'
    assert test_matches == ref_matches


@pytest.mark.parametrize(
    "k, input_mask, ref_output_mask",
    [
        (
            3,
            [0,1,1,0,0,0,0,0,1],
            [0,0,0,1,1,1,0]
        ),
        (
            4,
            [1,0,0,0,0,1,0,0,0,0,0,1,0,0],
            [0,1,0,0,0,0,1,1,0,0,0]
        ),
    ]
)
def test_masking(k, input_mask, ref_output_mask):
    """
    Explicitly test the conversion of removal masks to k-mer masks
    using known examples.
    Since the conversion function is private, this is tested indirectly,
    by looking at the sequence positions, that were added to the array.
    """
    input_mask = np.array(input_mask, dtype=bool)
    ref_output_mask = np.array(ref_output_mask, dtype=bool)

    sequence = seq.NucleotideSequence()
    sequence.code = np.zeros(len(input_mask))
    table = align.KmerTable.from_sequences(
        k, [sequence], ignore_masks=[input_mask]
    )

    # Get the k-mer positions that were masked
    test_output_mask = np.zeros(len(ref_output_mask), dtype=bool)
    for kmer in table.get_kmers():
        seq_indices = table[kmer][:, 1]
        test_output_mask[seq_indices] = True
    
    assert test_output_mask.tolist() == ref_output_mask.tolist()


def test_get_kmers():
    """
    Test whether the correct used *k-mers* are returned by
    :meth:`get_kmer()`, by constructing a table with exactly one
    appearance for each *k-mer* in a random list of *k-mers*.
    """
    np.random.seed(0)

    kmer_alphabet = align.KmerAlphabet(
        seq.NucleotideSequence.unambiguous_alphabet(),
        8
    )
    ref_mask = np.random.choice([False, True], size=len(kmer_alphabet))
    ref_kmers = np.where(ref_mask)[0]
    kmer_dict = {kmer: np.zeros((1, 2), dtype=np.uint32) for kmer in ref_kmers}
    # [[0, 0]] = Dummy position
    # The actual position is not relevant for the tested function
    table = align.KmerTable.from_positions(kmer_alphabet, kmer_dict)

    test_kmers = table.get_kmers()

    assert test_kmers.tolist() == ref_kmers.tolist()


def test_pickle(alphabet, k, random_sequences):
    """
    Test support of the pickle protocol by pickling and unpickling
    a :class:`KmerTable` and checking if the object is still equal to
    the original one.
    """
    ref_table = align.KmerTable.from_sequences(k, random_sequences)
    
    test_table = pickle.loads(pickle.dumps(ref_table))

    assert test_table == ref_table


def _identity_rule(alphabet):
    score_matrix = np.full((len(alphabet),) * 2, -1, dtype=int)
    np.fill_diagonal(score_matrix, 0)
    matrix = align.SubstitutionMatrix(alphabet, alphabet, score_matrix)
    rule = align.ScoreThresholdRule(matrix, 0)
    return rule


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import biotite.structure as struc
import numpy as np
import pytest


@pytest.fixture
def stack():
    stack = struc.AtomArrayStack(depth=3, length=5)
    stack.coord = np.arange(45).reshape((3,5,3))
    return stack


@pytest.mark.parametrize("as_coord", [False, True])
def test_rmsd(stack, as_coord):
    if as_coord:
        stack = stack.coord
    assert struc.rmsd(stack[0], stack).tolist() \
           == pytest.approx([0.0, 25.98076211, 51.96152423])
    assert struc.rmsd(stack[0], stack[1]) \
            == pytest.approx(25.9807621135)


@pytest.mark.parametrize("as_coord", [False, True])
def test_rmsf(stack, as_coord):
    if as_coord:
        stack = stack.coord
    assert struc.rmsf(struc.average(stack), stack).tolist() \
           == pytest.approx([21.21320344] * 5)


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information..

"""
This module contains all possible errors of the `structure` subpackage.
"""

__name__ = "biotite.structure"
__author__ = "Patrick Kunzmann"
__all__ = ["BadStructureError", "IncompleteStructureWarning",
           "UnexpectedStructureWarning"]


class BadStructureError(Exception):
    """
    Indicates that a structure is not suitable for a certain operation.
    """
    pass

class IncompleteStructureWarning(Warning):
    """
    Indicates that a structure is not complete.
    """
    pass

class UnexpectedStructureWarning(Warning):
    """
    Indicates that a structure was not expected.
    """
    pass



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import pickle
import numpy as np
import pytest
import biotite.structure as struc


@pytest.fixture
def atom_list():
    chain_id = ["A","A","B","B","B"]
    res_id = [1,1,1,1,2]
    ins_code = [""] * 5
    res_name = ["ALA","ALA","PRO","PRO","MSE"]
    hetero = [False, False, False, False, True]
    atom_name = ["N", "CA", "O", "CA", "SE"]
    element = ["N","C","O","C","SE"]
    atom_list = []
    for i in range(5):
        atom_list.append(struc.Atom([i,i,i],
                         chain_id = chain_id[i],
                         res_id = res_id[i],
                         ins_code = ins_code[i],
                         res_name = res_name[i],
                         hetero = hetero[i],
                         atom_name = atom_name[i],
                         element = element[i]))
    return atom_list

@pytest.fixture
def atom(atom_list):
    return atom_list[2]

@pytest.fixture
def array(atom_list):
    return struc.array(atom_list)

@pytest.fixture
def stack(array):
    return struc.stack([array, array.copy(), array.copy()])

@pytest.fixture
def array_box():
    return np.array([
        [1,0,0],
        [0,2,0],
        [0,0,3]
    ])

@pytest.fixture
def stack_box(stack, array_box):
    return np.array([array_box] * stack.stack_depth())

def test_shape(array, stack):
    assert array.shape == (5,)
    assert stack.shape == (3, 5)

def test_access(array):
    chain_id = ["A","A","B","B","B"]
    assert array.coord.shape == (5,3)
    assert array.chain_id.tolist() == chain_id
    assert array.get_annotation("chain_id").tolist() == chain_id
    array.add_annotation("test1", dtype=int)
    assert array.test1.tolist() == [0,0,0,0,0]
    with pytest.raises(IndexError):
        array.set_annotation("test2", np.array([0,1,2,3]))


def test_modification(atom, array, stack):
    new_atom = atom
    new_atom.chain_id = "C"
    del array[2]
    assert array.chain_id.tolist() == ["A","A","B","B"]
    array[-1] = new_atom
    assert array.chain_id.tolist() == ["A","A","B","C"]
    del stack[1]
    assert stack.stack_depth() == 2


def test_array_indexing(atom, array):
    filtered_array = array[array.chain_id == "B"]
    assert filtered_array.res_name.tolist() == ["PRO","PRO","MSE"] 
    assert atom == filtered_array[0]
    filtered_array = array[[0,2,4]]
    assert filtered_array.element.tolist() == ["N","O","SE"]


def test_stack_indexing(stack):
    with pytest.raises(IndexError):
        stack[5]
    filtered_stack = stack[0]
    assert type(filtered_stack) == struc.AtomArray
    filtered_stack = stack[0:2, stack.res_name == "PRO"]
    assert filtered_stack.atom_name.tolist() == ["O","CA"]
    filtered_stack = stack[np.array([True,False,True])]
    assert filtered_stack.stack_depth() == 2
    assert filtered_stack.array_length() == 5
    filtered_stack = stack[:,0]
    assert filtered_stack.stack_depth() == 3
    assert filtered_stack.array_length() == 1
    

def test_concatenation(array, stack):
    concat_array = array[2:] + array[:2]
    assert concat_array.chain_id.tolist() == ["B","B","B","A","A"]
    assert concat_array.coord.shape == (5,3)
    concat_stack = stack[:,2:] + stack[:,:2]
    assert concat_array.chain_id.tolist() == ["B","B","B","A","A"]
    assert concat_stack.coord.shape == (3,5,3)


def test_comparison(array):
    mod_array = array.copy()
    assert mod_array == array
    mod_array.coord += 1
    assert mod_array != array
    mod_array = array.copy()
    mod_array.res_name[0] = "UNK"
    mod_array != array


def test_bonds(array):
    assert array.bonds is None
    with pytest.raises(TypeError):
        # Expect a BondList
        array.bonds = 42
    with pytest.raises(ValueError):
        # Expect a BondList with array length as atom count
        array.bonds = struc.BondList(13)
    array.bonds = struc.BondList(5, np.array([(0,1),(0,2),(2,3),(2,4)]))
    assert array.bonds.as_array().tolist() == [[0, 1, 0],
                                               [0, 2, 0],
                                               [2, 3, 0],
                                               [2, 4, 0],]
    filtered_array = array[array.chain_id == "B"]
    assert filtered_array.bonds.as_array().tolist() == [[0, 1, 0],
                                                        [0, 2, 0]]
    concat_array = array + array
    assert concat_array.bonds.as_array().tolist() == [[0, 1, 0],
                                                      [0, 2, 0],
                                                      [2, 3, 0],
                                                      [2, 4, 0],
                                                      [5, 6, 0],
                                                      [5, 7, 0],
                                                      [7, 8, 0],
                                                      [7, 9, 0]]


def test_box(array, stack, array_box, stack_box):
    # Test attribute access
    with pytest.raises(ValueError):
        array.box = stack_box
    with pytest.raises(ValueError):
        stack.box = array_box
    with pytest.raises(ValueError):
        array.box = np.array([42])
    with pytest.raises(ValueError):
        stack.box = np.array([42])
    array.box = array_box
    stack.box = stack_box
    # Test indexing
    assert (stack[0].box == array_box).all()
    assert (stack[:2].box == np.array([array_box] * 2)).all()
    assert (stack[:2, 3].box == np.array([array_box] * 2)).all()
    assert (stack[[True, False, True]].box == np.array([array_box] * 2)).all()


def test_array_from_atoms(atom_list):
    """
    Check whether custom annotations in :class:`Atom` objects are
    properly carried over to the :class:`AtomArray` when using
    :func:`array()`.
    """
    for atom in atom_list:
        atom.some_annotation = 42
    array = struc.array(atom_list)
    assert np.all(array.some_annotation == np.full(array.array_length(), 42))
    assert np.issubdtype(array.some_annotation.dtype, np.integer)


def test_pickle(atom, array, stack):
    """
    Check if pickling and unpickling works.
    This test is necessary since the classes implement the
    :meth:`__getattr__()` and :meth:`__setattr__()` methods.
    """
    test_atom = pickle.loads(pickle.dumps(atom))
    assert test_atom == atom

    test_array = pickle.loads(pickle.dumps(array))
    assert test_array == array

    test_stack = pickle.loads(pickle.dumps(stack))
    assert test_stack == stack


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.sequence.io.fasta"
__author__ = "Patrick Kunzmann"
__all__ = ["FastaFile"]

from ....file import TextFile, InvalidFileError, wrap_string
from collections import OrderedDict
from collections.abc import MutableMapping


class FastaFile(TextFile, MutableMapping):
    """
    This class represents a file in FASTA format.
    
    A FASTA file contains so called *header* lines, beginning with
    ``>``, that describe following sequence.
    The corresponding sequence starts at the line after the header line
    and ends at the next header line or at the end of file.
    The header along with its sequence forms an entry.
    
    This class is used in a dictionary like manner, implementing the
    :class:`MutableMapping` interface:
    Headers (without the leading ``>``) are used as keys,
    and strings containing the sequences are the corresponding values.
    Entries can be accessed using indexing,
    ``del`` deletes the entry at the given index.

    Parameters
    ----------
    chars_per_line : int, optional
        The number characters in a line containing sequence data
        after which a line break is inserted.
        Only relevant, when adding sequences to a file.
        Default is 80.
    
    Examples
    --------
    
    >>> import os.path
    >>> file = FastaFile()
    >>> file["seq1"] = "ATACT"
    >>> print(file["seq1"])
    ATACT
    >>> file["seq2"] = "AAAATT"
    >>> print(file)
    >seq1
    ATACT
    >seq2
    AAAATT
    >>> print(dict(file.items()))
    {'seq1': 'ATACT', 'seq2': 'AAAATT'}
    >>> for header, seq in file.items():
    ...     print(header, seq)
    seq1 ATACT
    seq2 AAAATT
    >>> del file["seq1"]
    >>> print(dict(file.items()))
    {'seq2': 'AAAATT'}
    >>> file.write(os.path.join(path_to_directory, "test.fasta"))
    """
    
    def __init__(self, chars_per_line=80):
        super().__init__()
        self._chars_per_line = chars_per_line
        self._entries = OrderedDict()
    
    @classmethod
    def read(cls, file, chars_per_line=80):
        """
        Read a FASTA file.
        
        Parameters
        ----------
        file : file-like object or str
            The file to be read.
            Alternatively a file path can be supplied.
        chars_per_line : int, optional
            The number characters in a line containing sequence data
            after which a line break is inserted.
            Only relevant, when adding sequences to a file.
            Default is 80.
        
        Returns
        -------
        file_object : FastaFile
            The parsed file.
        """
        file = super().read(file, chars_per_line)
        # Filter out empty and comment lines
        file.lines = [line for line in file.lines
                      if len(line.strip()) != 0 and line[0] != ";"]
        if len(file.lines) == 0:
            raise InvalidFileError("File is empty or contains only comments")
        file._find_entries()
        return file
        
    def __setitem__(self, header, seq_str):
        if not isinstance(header, str):
            raise IndexError(
                "'FastaFile' only supports header strings as keys"
            )
        if not isinstance(seq_str, str):
            raise TypeError("'FastaFile' only supports sequence strings "
                             "as values")
        # Create lines for new header and sequence (with line breaks)
        new_lines = [">" + header.replace("\n","").strip()] + \
                    wrap_string(seq_str, width=self._chars_per_line)
        if header in self:
            # Delete lines of entry corresponding to the header,
            # if existing
            del self[header]
            self.lines += new_lines
            self._find_entries()
        else:
            # Simply append lines
            # Add entry in a more efficient way than '_find_entries()'
            # for this simple case
            self._entries[header] = (
                len(self.lines),
                len(self.lines) + len(new_lines)
            )
            self.lines += new_lines
    
    def __getitem__(self, header):
        if not isinstance(header, str):
            raise IndexError(
                "'FastaFile' only supports header strings as keys"
            )
        start, stop = self._entries[header]
        # Concatenate sequence string from following lines
        seq_string = "".join(
            [line.strip() for line in self.lines[start+1 : stop]]
        )
        return seq_string
    
    def __delitem__(self, header):
        start, stop = self._entries[header]
        del self.lines[start:stop]
        del self._entries[header]
        self._find_entries()
    
    def __len__(self):
        return len(self._entries)
    
    def __iter__(self):
        return self._entries.__iter__()
    
    def __contains__(self, identifer):
        return identifer in self._entries
    
    def _find_entries(self):
        if len(self.lines) > 0 and self.lines[0][0] != ">":
            raise InvalidFileError(
                f"File starts with '{self.lines[0][0]}' instead of '>'"
            )
        
        header_i = []
        for i, line in enumerate(self.lines):
            if line[0] == ">":
                header_i.append(i)
        
        self._entries = OrderedDict()
        for j in range(len(header_i)):
            # Remove leading '>' from header
            header = self.lines[header_i[j]].strip()[1:]
            start = header_i[j]
            if j < len(header_i) -1:
                # Header in mid or start of file
                # -> stop is start of next header
                stop = header_i[j+1]
            else:
                # Last header -> entry stops at end of file
                stop = len(self.lines)
            self._entries[header] = (start, stop)


    @staticmethod
    def read_iter(file):
        """
        Create an iterator over each sequence of the given FASTA file.
        
        Parameters
        ----------
        file : file-like object or str
            The file to be read.
            Alternatively a file path can be supplied.
        
        Yields
        ------
        header : str
            The header of the current sequence.
        seq_str : str
            The current sequence as string.
        
        Notes
        -----
        This approach gives the same results as
        `FastaFile.read(file).items()`, but is slightly faster and much
        more memory efficient.
        """
        header = None
        seq_str_list = []
        for line in TextFile.read_iter(file):
            line = line.strip()
            # Ignore empty and comment lines
            if len(line) == 0 or line[0] == ";":
                continue
            if line[0] == ">":
                # New entry
                # -> yield previous entry
                if header is not None:
                    yield header, "".join(seq_str_list)
                # Track new header and reset sequence
                header = line[1:]
                seq_str_list = []
            else:
                seq_str_list.append(line)
        # Yield final entry
        if header is not None:
            yield header, "".join(seq_str_list)
    

    @staticmethod
    def write_iter(file, items, chars_per_line=80):
        """
        Iterate over the given `items` and write each item into
        the specified `file`.

        In contrast to :meth:`write()`, the lines of text are not stored
        in an intermediate :class:`TextFile`, but are directly written
        to the file.
        Hence, this static method may save a large amount of memory if
        a large file should be written, especially if the `items`
        are provided as generator.
        
        Parameters
        ----------
        file : file-like object or str
            The file to be written to.
            Alternatively a file path can be supplied.
        items : generator or array-like of tuple(str, str)
            The entries to be written into the file.
            Each entry consists of an header string and a sequence
            string.
        chars_per_line : int, optional
            The number characters in a line containing sequence data
            after which a line break is inserted.
            Only relevant, when adding sequences to a file.
            Default is 80.

        Notes
        -----
        This method does not test, whether the given identifiers are
        unambiguous.
        """
        def line_generator():
            for item in items:
                header, seq_str = item
                if not isinstance(header, str):
                    raise IndexError(
                        "'FastaFile' only supports header strings"
                    )
                if not isinstance(seq_str, str):
                    raise TypeError(
                        "'FastaFile' only supports sequence strings"
                    )
                
                # Yield header line
                yield ">" + header.replace("\n","").strip()

                # Yield sequence line(s)
                for line in wrap_string(seq_str, width=chars_per_line):
                    yield line
    
        TextFile.write_iter(file, line_generator())


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

from os.path import dirname, abspath, join
import glob
import pytest
import biotite.sequence as seq
from ..util import cannot_import


@pytest.mark.skipif(
    cannot_import("matplotlib"), reason="Matplotlib is not installed"
)
@pytest.mark.parametrize(
    "scheme_path", glob.glob(
        join(
            dirname(abspath(seq.__file__)),
            "graphics", "color_schemes", "*.json"
        )
    )
)
def test_load_color_scheme(scheme_path):
    from matplotlib.colors import to_rgb
    import biotite.sequence.graphics as graphics

    supported_alphabets = [
        seq.NucleotideSequence.alphabet_amb,
        seq.ProteinSequence.alphabet,
        seq.LetterAlphabet("abcdefghijklmnop") # Protein block alphabet
    ]
    
    test_scheme = graphics.load_color_scheme(scheme_path)

    assert test_scheme["alphabet"] in supported_alphabets
    assert len(test_scheme["colors"]) == len(test_scheme["alphabet"])
    for color in test_scheme["colors"]:
        if color is not None:
            # Should not raise error
            to_rgb(color)


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import itertools
import datetime
from tempfile import TemporaryFile
import glob
from os.path import join, split, splitext
import pytest
import numpy as np
import biotite.structure.info as info
import biotite.structure.io.mol as mol
from ..util import data_dir


def test_header_conversion():
    """
    Write known example data to the header of a MOL file and expect
    to retrieve the same information when reading the file again.
    """
    ref_header = (
        "TestMol", "JD", "Biotite",
        datetime.datetime.now().replace(second=0, microsecond=0),
        "3D", "Lorem", "Ipsum", "123", "Lorem ipsum dolor sit amet"
    )

    mol_file = mol.MOLFile()
    mol_file.set_header(*ref_header)
    print(mol_file)
    temp = TemporaryFile("w+")
    mol_file.write(temp)

    temp.seek(0)
    mol_file = mol.MOLFile.read(temp)
    test_header = mol_file.get_header()
    temp.close()

    assert test_header == ref_header


@pytest.mark.parametrize(
    "path, omit_charge",
    itertools.product(
        glob.glob(join(data_dir("structure"), "molecules", "*.sdf")),
        [False, True]
    )
)
def test_structure_conversion(path, omit_charge):
    """
    After reading a MOL file, writing the structure back to a new file
    and reading it again should give the same structure.

    In this case an SDF file is used, but it is compatible with the
    MOL format.
    """
    mol_file = mol.MOLFile.read(path)
    ref_atoms = mol.get_structure(mol_file)
    print(ref_atoms.charge)
    if omit_charge:
        ref_atoms.del_annotation("charge")

    mol_file = mol.MOLFile()
    mol.set_structure(mol_file, ref_atoms)
    temp = TemporaryFile("w+")
    mol_file.write(temp)

    temp.seek(0)
    mol_file = mol.MOLFile.read(temp)
    test_atoms = mol.get_structure(mol_file)
    if omit_charge:
        assert np.all(test_atoms.charge == 0)
        test_atoms.del_annotation("charge") 
    temp.close()

    assert test_atoms == ref_atoms


@pytest.mark.parametrize(
    "path", glob.glob(join(data_dir("structure"), "molecules", "*.sdf")),
)
def test_pdbx_consistency(path):
    """
    Check if the structure parsed from a MOL file is equal to the same
    structure read from the *Chemical Component Dictionary* in PDBx
    format.

    In this case an SDF file is used, but it is compatible with the
    MOL format.
    """
    mol_name = split(splitext(path)[0])[1]
    ref_atoms = info.residue(mol_name)
    # The CCD contains information about aromatic bond types,
    # but the SDF test files do not
    ref_atoms.bonds.remove_aromaticity()

    mol_file = mol.MOLFile.read(path)
    test_atoms = mol_file.get_structure()

    assert test_atoms.coord.shape == ref_atoms.coord.shape
    assert test_atoms.coord.flatten().tolist() \
        == ref_atoms.coord.flatten().tolist()
    assert test_atoms.element.tolist() == ref_atoms.element.tolist()
    assert test_atoms.charge.tolist() == ref_atoms.charge.tolist()
    assert set(tuple(bond) for bond in test_atoms.bonds.as_array()) \
        == set(tuple(bond) for bond in  ref_atoms.bonds.as_array())


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This is the top-level package of *Biotite*.
Although it does not provide useful functionality for most users,
it does provide utilities and base classes used by a lot of *Biotite*'s
modules.
"""

__version__ = "0.33.0"
__name__ = "biotite"
__author__ = "Patrick Kunzmann"

from .file import *
from .temp import *
from .copyable import * 
from .visualize import * 



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.application.sra"
__author__ = "Patrick Kunzmann"
__all__ = ["FastqDumpApp"]

import glob
from tempfile import NamedTemporaryFile, gettempdir
from ..localapp import LocalApp, cleanup_tempfile
from ..application import AppState, requires_state
from ...sequence.io.fastq.file import FastqFile
from ...sequence.io.fastq.convert import get_sequences


class FastqDumpApp(LocalApp):
    """
    Fetch sequencing data as FASTQ from the *NCBI sequence read archive*
    (SRA) using *sra-tools*.

    Parameters
    ----------
    uid : str
        A *unique identifier* (UID) of the file to be downloaded.
    output_path_prefix : str, optional
        The prefix of the path to store the downloaded FASTQ file.
        ``.fastq`` is appended to this prefix if the run contains
        a single read per spot.
        ``_1.fastq``, ``_2.fastq``, etc. is appended if it contains
        multiple reads per spot.
        By default, the files are created in a temporary directory and
        deleted after the files have been read.
    bin_path : str, optional
        Path to the ``fasterq-dump`` binary.
    offset : int or {'Sanger', 'Solexa', 'Illumina-1.3', 'Illumina-1.5', 'Illumina-1.8'}, optional
        This value is subtracted from the FASTQ ASCII code to obtain the
        quality score.
        Can either be directly the value, or a string that indicates
        the score format.
    """
    
    def __init__(self, uid, output_path_prefix=None, bin_path="fasterq-dump",
                 offset="Sanger"):
        super().__init__(bin_path)
        self._uid = uid
        self._offset = offset
        if output_path_prefix is None:
            # NamedTemporaryFile is only created to obtain prefix
            # for FASTQ files
            self._out_file = NamedTemporaryFile("r")
            self._prefix = self._out_file.name
        else:
            self._out_file = None
            self._prefix = output_path_prefix

    def run(self):
        self.set_arguments([
            "-o", self._prefix + ".fastq",
            "-t", gettempdir(),
            "-f",
            self._uid
        ])
        super().run()
    
    def evaluate(self):
        super().evaluate()
        self._file_names = (
            # For entries with one read per spot
            glob.glob(self._prefix +   ".fastq") + 
            # For entries with multiple reads per spot
            glob.glob(self._prefix + "_*.fastq")
        )
        # Only load FASTQ files into memory when needed
        self._fastq_files = None
    
    def clean_up(self):
        super().clean_up()
        if self._out_file is not None:
            # This file was only created to reserve a unique file name
            # Now it is not needed anymore
            self._out_file.close()
    
    @requires_state(AppState.JOINED)
    def get_file_paths(self):
        """
        Get the file paths to the downloaded FASTQ files.
        
        Returns
        -------
        paths : list of str
            The file paths to the downloaded files.
        """
        return self._file_names
    
    @requires_state(AppState.JOINED)
    def get_sequences(self):
        """
        Get the sequences and score values from the downloaded file(s).
        
        Returns
        -------
        sequences_and_scores : list of dict (str -> (NucleotideSequence, ndarray))
            This list contains the reads for each spot:
            The first item contains the first read for each spot, the
            second item contains the second read for each spot (if existing),
            etc.
            Each item in the list is a dictionary mapping identifiers to its
            corresponding sequence and score values.
        """
        fastq_files = self.get_fastq()
        return [get_sequences(fastq_file) for fastq_file in fastq_files]

    @requires_state(AppState.JOINED)
    def get_fastq(self):
        """
        Get the `FastqFile` objects from the downloaded file(s).

        Returns
        -------
        fastq_files : list of FastqFile
            This list contains the reads for each spot:
            The first item contains the first read for each spot, the
            second item contains the second read for each spot (if existing),
            etc.
        """
        if self._fastq_files is None:
            self._fastq_files = [
                FastqFile.read(file_name, offset=self._offset)
                for file_name in self.get_file_paths()
            ]
        return self._fastq_files
    
    @staticmethod
    def fetch(uid, output_path_prefix=None, bin_path="fasterq-dump",
              offset="Sanger"):
        """
        Get the sequences and score values belonging to the UID from the
        *NCBI sequence read archive* (SRA).

        Parameters
        ----------
        uid : str
            A *unique identifier* (UID) of the file to be downloaded.
        output_path_prefix : str, optional
            The prefix of the path to store the downloaded FASTQ file.
            ``.fastq`` is appended to this prefix if the run contains
            a single read per spot.
            ``_1.fastq``, ``_2.fastq``, etc. is appended if it contains
            multiple reads per spot.
            By default, the files are created in a temporary directory and
            deleted after the files have been read.
        bin_path : str, optional
            Path to the ``fasterq-dump`` binary.
        offset : int or {'Sanger', 'Solexa', 'Illumina-1.3', 'Illumina-1.5', 'Illumina-1.8'}, optional
            This value is subtracted from the FASTQ ASCII code to obtain the
            quality score.
            Can either be directly the value, or a string that indicates
            the score format.
        
        Returns
        -------
        sequences_and_scores : list of dict (str -> (NucleotideSequence, ndarray))
            This list contains the reads for each spot:
            The first item contains the first read for each spot, the
            second item contains the second read for each spot (if existing),
            etc.
            Each item in the list is a dictionary mapping identifiers to its
            corresponding sequence and score values.
        """
        app = FastqDumpApp(uid, output_path_prefix, bin_path, offset)
        app.start()
        app.join()
        return app.get_sequences()



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite"
__author__ = "Patrick Kunzmann"
__all__ = ["temp_file", "temp_dir"]

import shutil
import atexit
import os
import tempfile
import warnings


_temp_dir = ""

def _create_temp_dir():
    global _temp_dir
    if _temp_dir == "":
        cwd = os.getcwd()
        _temp_dir = os.path.join(cwd, ".biotitetemp")
        if not os.path.isdir(_temp_dir):
            os.makedirs(_temp_dir)
        atexit.register(_delete_temp)


def _delete_temp():
    global _temp_dir
    # Condition only for savety reasons
    if ".biotitetemp" in _temp_dir: 
        shutil.rmtree(_temp_dir)


def temp_file(suffix=""):
    """
    Get a file path to a temporary file.
    
    All temporary files will be deleted after script execution.

    DEPRECATED: Please use the :mod:`tempfile` module from the
    standard library.
    
    Parameters
    ----------
    suffix : str
        Suffix of the file.
        By default no suffix will be appended.
    
    Returns
    -------
    temp_file_name : str
        a file name in the temporary directory.
    """
    global _temp_dir
    warnings.warn(
        "Please use the :mod:`tempfile` module from the standard library",
        DeprecationWarning
    )
    _create_temp_dir()
    if suffix != "" and not suffix.startswith("."):
        suffix = "." + suffix
    return tempfile.mktemp(suffix=suffix, dir=_temp_dir)
     

def temp_dir():
    """
    Get the temporary directory path.
    
    The temporary directory will be deleted after script execution.

    DEPRECATED: Please use the :mod:`tempfile` module from the
    standard library.
    
    Returns
    -------
    temp_dir : str
        Path of the temporary directory.
    """
    global _temp_dir
    warnings.warn(
        "Please use the :mod:`tempfile` module from the standard library",
        DeprecationWarning
    )
    _create_temp_dir()
    return _temp_dir


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.application.clustalo"
__author__ = "Patrick Kunzmann"
__all__ = ["ClustalOmegaApp"]

from tempfile import NamedTemporaryFile
import numpy as np
from ...sequence.sequence import Sequence
from ...sequence.seqtypes import NucleotideSequence, ProteinSequence
from ...sequence.io.fasta.file import FastaFile
from ...sequence.align.alignment import Alignment
from ...sequence.phylo.tree import Tree
from ..localapp import cleanup_tempfile
from ..msaapp import MSAApp
from ..application import AppState, requires_state


class ClustalOmegaApp(MSAApp):
    """
    Perform a multiple sequence alignment using Clustal-Omega.
    
    Parameters
    ----------
    sequences : list of ProteinSequence or NucleotideSequence
        The sequences to be aligned.
    bin_path : str, optional
        Path of the Custal-Omega binary.
    matrix : None
        This parameter is used for compatibility reasons and is ignored.
    
    Examples
    --------

    >>> seq1 = ProteinSequence("BIQTITE")
    >>> seq2 = ProteinSequence("TITANITE")
    >>> seq3 = ProteinSequence("BISMITE")
    >>> seq4 = ProteinSequence("IQLITE")
    >>> app = ClustalOmegaApp([seq1, seq2, seq3, seq4])
    >>> app.start()
    >>> app.join()
    >>> alignment = app.get_alignment()
    >>> print(alignment)
    -BIQTITE
    TITANITE
    -BISMITE
    --IQLITE
    """
    
    def __init__(self, sequences, bin_path="clustalo", matrix=None):
        super().__init__(sequences, bin_path, None)
        self._seq_count = len(sequences)
        self._mbed = True
        self._dist_matrix = None
        self._tree = None
        self._in_dist_matrix_file = NamedTemporaryFile(
            "w", suffix=".mat", delete=False
        )
        self._out_dist_matrix_file = NamedTemporaryFile(
            "r", suffix=".mat", delete=False
        )
        self._in_tree_file = NamedTemporaryFile(
            "w", suffix=".tree", delete=False
        )
        self._out_tree_file = NamedTemporaryFile(
            "r", suffix=".tree", delete=False
        )
    
    def run(self):
        args = [
            "--in", self.get_input_file_path(),
            "--out", self.get_output_file_path(),
            # The temporary files are already created
            # -> tell Clustal to overwrite these empty files
            "--force",
            # Tree order for get_alignment_order() to work properly 
            "--output-order=tree-order",
        ]
        if self.get_seqtype() == "protein":
            args += ["--seqtype", "Protein"]
        else:
            args += ["--seqtype", "DNA"]
        if self._tree is None:
            # ClustalOmega does not like when a tree is set
            # as input and output#
            # -> Only request tree output when not tree is input
            args += [
                "--guidetree-out", self._out_tree_file.name,
            ]
        if not self._mbed:
            args += [
                "--full",
                "--distmat-out", self._out_dist_matrix_file.name
            ]
        if self._dist_matrix is not None:
            # Add the sequence names (0, 1, 2, 3 ...) as first column
            dist_matrix_with_index = np.concatenate(
                (
                    np.arange(self._seq_count)[:, np.newaxis],
                    self._dist_matrix
                ), axis=1
            )
            np.savetxt(
                self._in_dist_matrix_file.name, dist_matrix_with_index,
                # The first line contains the amount of sequences
                comments = "",
                header = str(self._seq_count),
                # The sequence indices are integers, the rest are floats
                fmt = ["%d"] + ["%.5f"] * self._seq_count
            )
            args += ["--distmat-in", self._in_dist_matrix_file.name]
        if self._tree is not None:
            self._in_tree_file.write(str(self._tree))
            self._in_tree_file.flush()
            args += ["--guidetree-in", self._in_tree_file.name]
        self.set_arguments(args)
        super().run()
    
    def evaluate(self):
        super().evaluate()
        if not self._mbed:
            self._dist_matrix = np.loadtxt(
                self._out_dist_matrix_file.name,
                # The first row only contains the number of sequences
                skiprows = 1,
                dtype = float
            )
            # The first column contains only the name of the
            # sequences, in this case 0, 1, 2, 3 ...
            # -> Omit the first column
            self._dist_matrix = self._dist_matrix[:, 1:]
        # Only read output tree if no tree was input
        if self._tree is None:
            self._tree = Tree.from_newick(
                self._out_tree_file.read().replace("\n", "")
            )
    
    def clean_up(self):
        super().clean_up()
        cleanup_tempfile(self._in_dist_matrix_file)
        cleanup_tempfile(self._out_dist_matrix_file)
        cleanup_tempfile(self._in_tree_file)
        cleanup_tempfile(self._out_tree_file)
    
    @requires_state(AppState.CREATED)
    def full_matrix_calculation(self):
        """
        Use full distance matrix for guide-tree calculation, equivalent
        to the ``--full`` option.

        This makes the distance matrix calculation slower than using the
        default *mBed* heuristic.
        """
        self._mbed = False
    
    @requires_state(AppState.CREATED)
    def set_distance_matrix(self, matrix):
        """
        Set the pairwise sequence distances, the program should use to
        calculate the guide tree. 
        
        Parameters
        ----------
        matrix : ndarray, shape=(n,n), dtype=float
            The pairwise distances.
        """
        if matrix.shape != (self._seq_count, self._seq_count):
            raise ValueError(
                f"Matrix with shape {matrix.shape} is not sufficient for "
                f"{self._seq_count} sequences"
            )
        self._dist_matrix = matrix.astype(float, copy=False)
    
    @requires_state(AppState.JOINED)
    def get_distance_matrix(self):
        """
        Get the pairwise sequence distances the program used to
        calculate the guide tree. 
        
        Returns
        -------
        matrix : ndarray, shape=(n,n), dtype=float
            The pairwise distances.
        """
        if self._mbed:
            raise ValueError(
                "Getting the distance matrix requires "
                "'full_matrix_calculation()'"
            )
        return self._dist_matrix
    
    @requires_state(AppState.CREATED)
    def set_guide_tree(self, tree):
        """
        Set the guide tree, the program should use for the
        progressive alignment.
        
        Parameters
        ----------
        tree : Tree
            The guide tree.
        """
        if self._seq_count != len(tree):
            raise ValueError(
                f"Tree with {len(tree)} leaves is not sufficient for "
                "{self._seq_count} sequences, must be equal"
            )
        self._tree = tree
    
    @requires_state(AppState.JOINED)
    def get_guide_tree(self):
        """
        Get the guide tree created for the progressive alignment.
        
        Returns
        -------
        tree : Tree
            The guide tree.
        """
        return self._tree
    
    @staticmethod
    def supports_nucleotide():
        return True
    
    @staticmethod
    def supports_protein():
        return True
    
    @staticmethod
    def supports_custom_nucleotide_matrix():
        return False
    
    @staticmethod
    def supports_custom_protein_matrix():
        return False



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

from tempfile import NamedTemporaryFile
from os.path import join
import numpy as np
import pytest
import biotite.structure as struc
from biotite.structure.io import load_structure, save_structure
from ..util import data_dir, cannot_import


# Ignore warning about dummy unit cell vector
@pytest.mark.filterwarnings("ignore")
@pytest.mark.skipif(
    cannot_import("mdtraj"),
    reason="MDTraj is not installed"
)
@pytest.mark.parametrize("pdb_id", ["1l2y", "1gya", "1igy"])
def test_hbond_structure(pdb_id):
    file_name = join(data_dir("structure"), pdb_id+".mmtf")
    
    array = load_structure(file_name)
    # Only consider amino acids for consistency
    # with bonded hydrogen detection in MDTraj
    array = array[..., struc.filter_amino_acids(array)]
    if isinstance(array, struc.AtomArrayStack):
        # For consistency with MDTraj 'S' cannot be acceptor element
        # https://github.com/mdtraj/mdtraj/blob/master/mdtraj/geometry/hbond.py#L365
        triplets, mask = struc.hbond(array, acceptor_elements=("O","N"))
    else:
        triplets = struc.hbond(array, acceptor_elements=("O","N"))
    
    # Save to new pdb file for consistent treatment of inscode/altloc
    # im MDTraj
    temp = NamedTemporaryFile("w+", suffix=".pdb")
    save_structure(temp.name, array)
    
    # Compare with MDTraj
    import mdtraj
    traj = mdtraj.load(temp.name)
    temp.close()
    triplets_ref = mdtraj.baker_hubbard(
        traj, freq=0, periodic=False
    )

    # Both packages may use different order
    # -> use set for comparison
    triplets_set = set([tuple(tri) for tri in triplets])
    triplets_ref_set = set([tuple(tri) for tri in triplets_ref])
    assert triplets_set == triplets_ref_set


def test_hbond_same_res():
    """
    Check if hydrogen bonds in the same residue are detected.
    At least one of such bonds is present in 1L2Y (1ASN with N-terminus)
    (model 2).
    """
    stack = load_structure(join(data_dir("structure"), "1l2y.mmtf"))
    selection = stack.res_id == 1
    # Focus on second model
    array = stack[1]
    triplets = struc.hbond(array, selection, selection)
    assert len(triplets) == 1


def test_hbond_total_count():
    """
    With the standart Baker & Hubbard criterion,
    1l2y should have 28 hydrogen bonds with a frequency > 0.1
    (comparision with MDTraj results)
    """
    stack = load_structure(join(data_dir("structure"), "1l2y.mmtf"))
    triplets, mask = struc.hbond(stack)
    freq = struc.hbond_frequency(mask)

    assert len(freq[freq >= 0.1]) == 28


def test_hbond_with_selections():
    """
    When selection1 and selection2 is defined, no hydrogen bonds outside
    of this boundary should be found. Also, hbond should respect the
    selection type.
    """
    stack = load_structure(join(data_dir("structure"), "1l2y.mmtf"))
    selection1 = (stack.res_id == 3) & (stack.atom_name == 'O')  # 3TYR BB Ox
    selection2 = stack.res_id == 7

    # backbone hbond should be found if selection1/2 type is both
    triplets, mask = struc.hbond(stack, selection1, selection2,
                                 selection1_type="both")
    assert len(triplets) == 1
    assert triplets[0][0] == 116
    assert triplets[0][2] == 38

    # backbone hbond should be found if selection1 is acceptor and
    # selection2 is donor
    triplets, mask = struc.hbond(stack, selection1, selection2,
                                 selection1_type="acceptor")
    assert len(triplets) == 1
    assert triplets[0][0] == 116
    assert triplets[0][2] == 38

    # no hbond should be found,
    # because the backbone oxygen cannot be a donor
    triplets, mask = struc.hbond(stack, selection1, selection2,
                                 selection1_type="donor")
    assert len(triplets) == 0


def test_hbond_single_selection():
    """
    If only selection1 or selection2 is defined, hbond should run
    against all other atoms as the other selection.
    """
    stack = load_structure(join(data_dir("structure"), "1l2y.mmtf"))
    selection = (stack.res_id == 2) & (stack.atom_name == "O")  # 2LEU BB Ox
    triplets, mask = struc.hbond(stack, selection1=selection)
    assert len(triplets) == 2

    triplets, mask = struc.hbond(stack, selection2=selection)
    assert len(triplets) == 2


def test_hbond_frequency():
    mask = np.array([
        [True, True, True, True, True], # 1.0
        [False, False, False, False, False], # 0.0
        [False, False, False, True, True] # 0.4
    ]).T
    freq = struc.hbond_frequency(mask)
    assert not np.isin(False, np.isclose(freq, np.array([1.0, 0.0, 0.4])))


@pytest.mark.parametrize("translation_vector", [(10,20,30), (-5, 3, 18)])
def test_hbond_periodicity(translation_vector):
    """
    Test whether hydrogen bond identification uses periodic boundary
    conditions correctly.
    For this purpose a structure containing water is loaded and the
    hydrogen bonds are identified.
    Then the position of the periodic boundary is changed and it is
    expected that all hydrogen bonds are still the same
    """
    stack = load_structure(join(data_dir("structure"), "waterbox.gro"))
    array = stack[0]
    ref_hbonds = struc.hbond(array, periodic=True)
    # Put H-bond triplets into as stack for faster comparison with
    # set for moved atoms
    ref_hbonds = set([tuple(triplet) for triplet in ref_hbonds])
    # Move system and put back into box
    # -> Equal to move of periodic boundary
    array = struc.translate(array, translation_vector)
    array.coord = struc.move_inside_box(array.coord, array.box)
    hbonds = struc.hbond(array, periodic=True)
    hbonds = set([tuple(triplet) for triplet in hbonds])
    assert ref_hbonds == hbonds


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
Functions for obtaining metadata fields of a GenBank file.
"""

__name__ = "biotite.sequence.io.genbank"
__author__ = "Patrick Kunzmann"
__all__ = ["get_locus", "get_definition", "get_accession", "get_version",
           "get_gi", "get_db_link", "get_source",
           "set_locus"]

from ....file import InvalidFileError
from .file import GenBankFile


def get_locus(gb_file):
    """
    Parse the *LOCUS* field of a GenBank or GenPept file.
    
    Parameters
    ----------
    gb_file : GenBankFile
        The GenBank file to read the *LOCUS* field from.

    Returns
    -------
    name : str
        The locus name.
    length : int
        Sequence length.
    mol_type : str, optional
        The molecule type.
        Usually one of ``'DNA'``, ``'RNA'``, ``'Protein'`` or ``''``.
    is_circular : bool, optional
        True, if the sequence is circular, false otherwise.
    division : str, optional
        The GenBank division to which the file belongs.
    date : str, optional
        The date of last modification.
    
    Examples
    --------
    
    >>> import os.path
    >>> file = GenBankFile.read(os.path.join(path_to_sequences, "ec_bl21.gb"))
    >>> name, length, mol_type, is_circular, division, date = get_locus(file)
    >>> print(name)
    CP001509
    >>> print(length)
    4558953
    >>> print(mol_type)
    DNA
    >>> print(is_circular)
    True
    >>> print(division)
    BCT
    >>> print(date)
    16-FEB-2017
    """
    lines, _ = _expect_single_field(gb_file, "LOCUS")
    # 'LOCUS' field has only one line
    locus_info = lines[0]

    name = locus_info[:18].strip()
    length = int(locus_info[18 : 28])
    mol_type = locus_info[32 : 42].strip()
    is_circular = True if locus_info[43 : 51] == "circular" else False
    division = locus_info[52 : 55].strip()
    date = locus_info[56 : 67]
    return name, length, mol_type, is_circular, division, date

def get_definition(gb_file):
    """
    Parse the *DEFINITION* field of a GenBank or GenPept file.
    
    Parameters
    ----------
    gb_file : GenBankFile
        The GenBank file to read the *DEFINITION* field from.
    
    Returns
    -------
    definition : str
        Content of the *DEFINITION* field.
    
    Examples
    --------
    
    >>> import os.path
    >>> file = GenBankFile.read(os.path.join(path_to_sequences, "ec_bl21.gb"))
    >>> print(get_definition(file))
    Escherichia coli BL21(DE3), complete genome.
    """
    lines, _ = _expect_single_field(gb_file, "DEFINITION")
    return " ".join([line.strip() for line in lines])

def get_accession(gb_file):
    """
    Parse the *ACCESSION* field of a GenBank or GenPept file.
    
    Parameters
    ----------
    gb_file : GenBankFile
        The GenBank file to read the *ACCESSION* field from.
    
    Returns
    -------
    accession : str
        The accession ID of the file.
    
    Examples
    --------
    
    >>> import os.path
    >>> file = GenBankFile.read(os.path.join(path_to_sequences, "ec_bl21.gb"))
    >>> print(get_accession(file))
    CP001509
    """
    lines, _ = _expect_single_field(gb_file, "ACCESSION")
    # 'ACCESSION' field has only one line
    return lines[0]

def get_version(gb_file):
    """
    Parse the version from the *VERSION* field of a GenBank or GenPept
    file.
    
    Parameters
    ----------
    gb_file : GenBankFile
        The GenBank file to read the *VERSION* field from.
    
    Returns
    -------
    version : str
        Content of the *VERSION* field. Does not include GI.
    """
    lines, _ = _expect_single_field(gb_file, "VERSION")
    # 'VERSION' field has only one line
    return lines[0].split()[0]

def get_gi(gb_file):
    """
    Parse the GI from the *VERSION* field of a GenBank or GenPept
    file.
    
    Parameters
    ----------
    gb_file : GenBankFile
        The GenBank file to read the *VERSION* field from.
    
    Returns
    -------
    gi : str
        The GI of the file.
    """
    lines, _ = _expect_single_field(gb_file, "VERSION")
    # 'VERSION' field has only one line
    version_info = lines[0].split()
    if len(version_info) < 2 or "GI" not in version_info[1]:
        raise InvalidFileError("File does not contain GI")
    # Truncate GI
    return int(version_info[1][3:])

def get_db_link(gb_file):
    """
    Parse the *DBLINK* field of a GenBank or GenPept file.
    
    Parameters
    ----------
    gb_file : GenBankFile
        The GenBank file to read the *DBLINK* field from.
    
    Returns
    -------
    link_dict : dict
        A dictionary storing the database links, with the database
        name as key, and the corresponding ID as value.
    
    Examples
    --------
    
    >>> import os.path
    >>> file = GenBankFile.read(os.path.join(path_to_sequences, "ec_bl21.gb"))
    >>> for key, val in get_db_link(file).items():
    ...     print(key, ":", val)
    BioProject : PRJNA20713
    BioSample : SAMN02603478
    """
    lines, _ = _expect_single_field(gb_file, "DBLINK")
    link_dict = {}
    for line in lines:
        key, value = line.split(":")
        link_dict[key.strip()] = value.strip()
    return link_dict


def get_source(gb_file):
    """
    Parse the *SOURCE* field of a GenBank or GenPept file.
    
    Parameters
    ----------
    gb_file : GenBankFile
        The GenBank file to read the *SOURCE* field from.
    
    Returns
    -------
    accession : str
        The name of the source organism.
    """
    lines, _ = _expect_single_field(gb_file, "SOURCE")
    # 'SOURCE' field has only one line
    return lines[0]


def _expect_single_field(gb_file, name):
    fields = gb_file.get_fields(name)
    if len(fields) == 0:
        raise InvalidFileError(f"File has no '{name}' field")
    if len(fields) > 1:
        raise InvalidFileError(f"File has multiple '{name}' fields")
    return fields[0]



def set_locus(gb_file, name, length, mol_type=None, is_circular=False,
              division=None, date=None):
    """
    Set the *LOCUS* field of a GenBank file.
    
    Parameters
    ----------
    gb_file : GenBankFile
        The GenBank file to be edited.
    name : str
        The locus name.
    length : int
        Sequence length.
    mol_type : str, optional
        The molecule type.
        Usually one of ``'DNA'``, ``'RNA'``, ``'Protein'`` or ``''``.
    is_circular : bool, optional
        True, if the sequence is circular, false otherwise.
    division : str, optional
        The GenBank division to which the file belongs.
    date : str, optional
        The date of last modification.
    """
    mol_type = "" if mol_type is None else mol_type
    restype_abbr = "aa" if mol_type in ["", "Protein"] else "bp"
    circularity = "circular" if is_circular else "linear"
    division = "" if division is None else division
    date = "" if date is None else date
    line = f"{name:18} {length:>9} {restype_abbr} {mol_type:^10} " \
           f"{circularity:8} {division:3} {date:11}"
    gb_file.set_field("LOCUS", [line])


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import pkgutil
from os.path import dirname, join, isdir, splitext
import importlib
import pytest


def find_all_modules(package_name, src_dir):
    """
    Recursively look for Python modules in the given source directory.
    (Sub-)Packages are not considered as modules.
    """
    module_names = []
    for _, module_name, is_package in pkgutil.iter_modules([src_dir]):
        full_module_name = f"{package_name}.{module_name}"
        if is_package:
            module_names.extend(find_all_modules(
                full_module_name,
                join(src_dir, module_name)
            ))
        else:
            module_names.append(full_module_name)
    return module_names


@pytest.mark.parametrize(
    "module_name",
    find_all_modules(
        "biotite",
        join(dirname(dirname(__file__)), "src", "biotite")
    )
)
def test_module_name(module_name):
    """
    Test whether the '__name__' attribute of each module in Biotite is
    set to the name of the subpackage.
    This needs to be tested, since by default the '__name__' attribute
    is equal to the name of the module.
    For example, we expect 'biotite.structure' instead of
    'biotite.structure.atoms'.
    """
    # Remove the part after the last '.' of the module name
    # to obtain the package name
    package_name = ".".join(module_name.split(".")[:-1])
    module = importlib.import_module(module_name)
    assert module.__name__ == package_name


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
A subpackage that provides interfaces for external software in case
*Biotite*’s integrated functionality is not sufficient for your tasks.
These interfaces range from locally installed software
(e.g. MSA software) to web services (e.g. BLAST).
The interfaces are seamless:
Writing input files and reading output files is handled internally.
The user only needs to provide objects like a :class:`Sequence`
and will receive objects like an :class:`Alignment`.

Note that in order to use an interface in :mod:`biotite.application`
the corresponding software must be installed or the web server must be
reachable, respectively.
These programs are not shipped with the *Biotite* package.

Each application is represented by its respective :class:`Application`
class.
:class:`Application` objects are created, started and after the run has
finished, the results are collected.
The current state of the the execution is indicated by an
:class:`AppState` object, which restricts which method calls are
allowed:
For example, the parameters can only be set, when the
:class:`Application` has not been started yet and the results can only
be collected after :class:`Application` has finished.

The execution of an :class:`Application` can run in parallel:
In the time between starting the run and collecting the results can be
used to run other code, similar to the *Python* :class:`Thread` or
:class:`Process` classes.
"""

__name__ = "biotite.application"
__author__ = "Patrick Kunzmann"

from .application import *
from .localapp import *
from .webapp import *
from .msaapp import *


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.application"
__author__ = "Patrick Kunzmann"
__all__ = ["map_sequence", "map_matrix"]


import numpy as np
from ..sequence.seqtypes import ProteinSequence
from ..sequence.align.matrix import SubstitutionMatrix


def map_sequence(sequence):
    """
    Map a sequence with an arbitrary alphabet into a
    :class:`ProteinSequence`, in order to support arbitrary sequence
    types in software that can handle protein sequences. 
    """
    if len(sequence.alphabet) > len(ProteinSequence.alphabet):
        # Cannot map into a protein sequence if the alphabet
        # has more symbols
        raise TypeError(
            f"The software cannot align sequences of type "
            f"{type(sequence).__name__}: "
            f"Alphabet is too large to be converted into amino "
            f"acid alphabet"
        )
    # Mapping is done by simply taking over the sequence
    # code of the original sequence
    mapped_sequence = ProteinSequence()
    mapped_sequence.code = sequence.code
    return mapped_sequence


def map_matrix(matrix):
    """
    Map a :class:`SubstitutionMatrix` with an arbitrary alphabet into a
    class:`SubstitutionMatrix` for protein sequences, in order to support
    arbitrary sequence types in software that can handle protein
    sequences. 
    """
    if matrix is None:
        raise TypeError(
            "A substitution matrix must be provided for custom "
            "sequence types"
        )
    # Create a protein substitution matrix with the values taken
    # from the original matrix
    # All trailing symbols are filled with zeros
    old_length = len(matrix.get_alphabet1())
    new_length = len(ProteinSequence.alphabet)
    new_score_matrix = np.zeros((new_length, new_length))
    new_score_matrix[:old_length, :old_length] = matrix.score_matrix()
    return SubstitutionMatrix(
        ProteinSequence.alphabet, ProteinSequence.alphabet,
        new_score_matrix
    )


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This module contains a convenience function for loading sequences from
general sequence files.
"""

__name__ = "biotite.sequence.io"
__author__ = "Patrick Kunzmann"
__all__ = ["load_sequence", "save_sequence",
           "load_sequences", "save_sequences"]

import itertools
import os.path
import io
from collections import OrderedDict
import numpy as np
from ..seqtypes import NucleotideSequence, ProteinSequence
from ..alphabet import Alphabet


def load_sequence(file_path):
    """
    Load a sequence from a sequence file without the need
    to manually instantiate a :class:`File` object.
    
    Internally this function uses a :class:`File` object, based on the
    file extension.
    
    Parameters
    ----------
    file_path : str
        The path to the sequence file.
    
    Returns
    -------
    sequence : Sequence
        The first sequence in the file.
    """
    # We only need the suffix here
    filename, suffix = os.path.splitext(file_path)
    if suffix in [".fasta", ".fa", ".mpfa", ".fna", ".fsa"]:
        from .fasta import FastaFile, get_sequence
        file = FastaFile.read(file_path)
        return get_sequence(file)
    elif suffix in [".fastq", ".fq"]:
        from .fastq import FastqFile
        # Quality scores are irrelevant for this function
        # -> Offset is irrelevant
        file = FastqFile.read(file_path, offset="Sanger")
        # Get first sequence
        for seq_str, scores in file.values():
            sequence = NucleotideSequence(seq_str)
            break
        return sequence
    elif suffix in [".gb", ".gbk", ".gp"]:
        from .genbank import GenBankFile, get_sequence
        format = "gp" if suffix == ".gp" else "gb"
        file = GenBankFile.read(file_path)
        return get_sequence(file, format)
    else:
        raise ValueError(f"Unknown file format '{suffix}'")


def save_sequence(file_path, sequence):
    """
    Save a sequence into a sequence file without the need
    to manually instantiate a :class:`File` object.
    
    Internally this function uses a :class:`File` object, based on the
    given file extension.
    
    Parameters
    ----------
    file_path : str
        The path to structure file.
    sequence : Sequence
        The sequence to be saved.
    """
    # We only need the suffix here
    filename, suffix = os.path.splitext(file_path)
    if suffix in [".fasta", ".fa", ".mpfa", ".fna", ".fsa"]:
        from .fasta import FastaFile, set_sequence
        file = FastaFile()
        set_sequence(file, sequence)
        file.write(file_path)
    elif suffix in [".fastq", ".fq"]:
        from .fastq import FastqFile
        # Quality scores are irrelevant for this function
        # -> Offset is irrelevant
        file = FastqFile(offset="Sanger")
        # Scores are set to 0 since no score information is supplied
        scores = np.zeros(len(sequence))
        file["sequence"] = str(sequence), scores
        file.write(file_path)
    elif suffix in [".gb", ".gbk", ".gp"]:
        from .genbank import GenBankFile, set_locus, set_sequence
        file = GenBankFile()
        set_locus(file, "sequence", len(sequence))
        set_sequence(file, sequence)
        file.write(file_path)
    else:
        raise ValueError(f"Unknown file format '{suffix}'")


def load_sequences(file_path):
    """
    Load multiple sequences from a sequence file without the need
    to manually instantiate a :class:`File` object.
    
    Internally this function uses a :class:`File` object, based on the
    file extension.
    
    Parameters
    ----------
    file_path : str
        The path to the sequence file.
    
    Returns
    -------
    sequences : dict of (str, Sequence)
        The sequences in the file.
        This dictionary maps each header name to
        the respective sequence. 
    """
    # We only need the suffix here
    filename, suffix = os.path.splitext(file_path)
    if suffix in [".fasta", ".fa", ".mpfa", ".fna", ".fsa"]:
        from .fasta import FastaFile, get_sequences
        file = FastaFile.read(file_path)
        return get_sequences(file)
    elif suffix in [".fastq", ".fq"]:
        from .fastq import FastqFile
        # Quality scores are irrelevant for this function
        # -> Offset is irrelevant
        file = FastqFile.read(file_path, offset="Sanger")
        return {identifier : NucleotideSequence(seq_str)
                for identifier, (seq_str, scores) in file.items()}
    elif suffix in [".gb", ".gbk", ".gp"]:
        from .genbank import MultiFile, get_definition, get_sequence
        file = MultiFile.read(file_path)
        format = "gp" if suffix == ".gp" else "gb"
        sequences = OrderedDict()
        for f in file:
            sequences[get_definition(f)] = get_sequence(f, format)
        return sequences
    else:
        raise ValueError(f"Unknown file format '{suffix}'")


def save_sequences(file_path, sequences):
    """
    Save multiple sequences into a sequence file without the need
    to manually instantiate a :class:`File` object.
    
    Internally this function uses a :class:`File` object, based on the
    given file extension.
    
    Parameters
    ----------
    file_path : str
        The path to structure file.
    sequences : dict of (str, Sequence)
        The sequences to be saved. The dictionary maps a header name
        to asequence.
    """
    # We only need the suffix here
    filename, suffix = os.path.splitext(file_path)
    if suffix in [".fasta", ".fa", ".mpfa", ".fna", ".fsa"]:
        from .fasta import FastaFile, set_sequences
        file = FastaFile()
        set_sequences(file, sequences)
        file.write(file_path)
    elif suffix in [".fastq", ".fq"]:
        from .fastq import FastqFile
        # Quality scores are irrelevant for this function
        # -> Offset is irrelevant
        file = FastqFile(offset="Sanger")
        for identifier, sequence in sequences.items():
            # Scores are set to 0 since no score information is supplied
            scores = np.zeros(len(sequence))
            file["identifer"] = str(sequence), scores
        file.write(file_path)
    elif suffix in [".gb", ".gbk", ".gp"]:
        raise NotImplementedError(
            "Writing GenBank files containing multiple records is currently "
            "not supported"
        )
    else:
        raise ValueError(f"Unknown file format '{suffix}'")



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.sequence.io.genbank"
__author__ = "Patrick Kunzmann"
__all__ = ["GenBankFile", "MultiFile"]

#import textwrap
import copy
#import re
import io
from ....file import TextFile, InvalidFileError
from collections import OrderedDict
#from ...annotation import Location, Feature, Annotation, AnnotatedSequence
#from ...seqtypes import NucleotideSequence, ProteinSequence 


class GenBankFile(TextFile):
    """
    This class represents a file in GenBank format (including GenPept).

    A GenBank file annotates a reference sequence with features such as
    positions of genes, promoters, etc.
    Additionally, it provides metadata further describing the file.

    A file is divided into separate fields, e.g. the *DEFINITION*
    field contains a description of the file.
    The field name starts at the beginning of a line,
    followed by the content.
    A field may contain subfields, whose name is indented.
    For example, the *SOURCE* field contains the *ORGANISM* subfield.
    Some fields may occur multiple times, e.g. the *REFERENCE* field.
    A sample GenBank file can be viewed at
    `<https://www.ncbi.nlm.nih.gov/Sitemap/samplerecord.html>`_.
    
    This class provides a low-level interface for parsing, editing and
    writing GenBank files.
    It works like a list of field entries, where a field consists of the
    field name, the field content and the subfields.
    The field content is separated into the lines belonging to the
    content.
    While the content of metadata fields starts at the standard
    GenBank indentation of 12, the content of the *FEATURES*
    (contains the annotation) and *ORIGIN* (contains the sequence)
    fields starts without indentation.
    The subfields are represented by a dictionary, with subfield names
    being keys and the corresponding lines being values.
    The *FEATURES* and *ORIGIN* fields have no subfields.
    
    Every entry can be obtained, set and deleted via the index operator.

    Notes
    -----
    This class does not support location identifiers with references
    to other Entrez database entries, e.g.
    ``join(1..100,J00194.1:100..202)``.
    
    Examples
    --------
    Create a GenBank file from scratch:

    >>> file = GenBankFile()
    >>> file.append(
    ...     "SOMEFIELD", ["One line", "A second line"],
    ...     subfields={"SUBFIELD1": ["Single Line"], "SUBFIELD2": ["Two", "lines"]}
    ... )
    >>> print(file)
    SOMEFIELD   One line
                A second line
      SUBFIELD1 Single Line
      SUBFIELD2 Two
                lines
    //
    >>> name, content, subfields = file[0]
    >>> print(name)
    SOMEFIELD
    >>> print(content)
    ['One line', 'A second line']
    >>> print(subfields)
    OrderedDict([('SUBFIELD1', ['Single Line']), ('SUBFIELD2', ['Two', 'lines'])])
    
    Adding an additional field:
    
    >>> file.insert(0, "OTHERFIELD", ["Another line"])
    >>> print(len(file))
    2
    >>> print(file)
    OTHERFIELD  Another line
    SOMEFIELD   One line
                A second line
      SUBFIELD1 Single Line
      SUBFIELD2 Two
                lines
    //

    Overwriting and deleting an existing field:

    >>> file[1] = "NEWFIELD", ["Yet another line"]
    >>> print(file)
    OTHERFIELD  Another line
    NEWFIELD    Yet another line
    //
    >>> file[1] = "NEWFIELD", ["Yet another line"], {"NEWSUB": ["Subfield line"]}
    >>> print(file)
    OTHERFIELD  Another line
    NEWFIELD    Yet another line
      NEWSUB    Subfield line
    //
    >>> del file[1]
    >>> print(file)
    OTHERFIELD  Another line
    //

    Parsing fields from a real GenBank file:

    >>> import os.path
    >>> file = GenBankFile.read(os.path.join(path_to_sequences, "gg_avidin.gb"))
    >>> print(file)
    LOCUS       AJ311647                1224 bp    DNA     linear   VRT 14-NOV-2006
    DEFINITION  Gallus gallus AVD gene for avidin, exons 1-4.
    ACCESSION   AJ311647
    VERSION     AJ311647.1  GI:13397825
    KEYWORDS    AVD gene; avidin.
    SOURCE      Gallus gallus (chicken)
      ORGANISM  Gallus gallus
                Eukaryota; Metazoa; Chordata; Craniata; Vertebrata; Euteleostomi;
                Archelosauria; Archosauria; Dinosauria; Saurischia; Theropoda;
                Coelurosauria; Aves; Neognathae; Galloanserae; Galliformes;
                Phasianidae; Phasianinae; Gallus.
    REFERENCE   1
      AUTHORS   Wallen,M.J., Laukkanen,M.O. and Kulomaa,M.S.
      TITLE     Cloning and sequencing of the chicken egg-white avidin-encoding
                gene and its relationship with the avidin-related genes Avr1-Avr5
      JOURNAL   Gene 161 (2), 205-209 (1995)
       PUBMED   7665080
    REFERENCE   2
      AUTHORS   Ahlroth,M.K., Kola,E.H., Ewald,D., Masabanda,J., Sazanov,A.,
                Fries,R. and Kulomaa,M.S.
      TITLE     Characterization and chromosomal localization of the chicken avidin
                gene family
      JOURNAL   Anim. Genet. 31 (6), 367-375 (2000)
       PUBMED   11167523
    REFERENCE   3  (bases 1 to 1224)
      AUTHORS   Ahlroth,M.K.
      TITLE     Direct Submission
      JOURNAL   Submitted (09-MAR-2001) Ahlroth M.K., Department of Biological and
                Environmental Science, University of Jyvaskyla, PO Box 35,
                FIN-40351 Jyvaskyla, FINLAND
    FEATURES             Location/Qualifiers
         source          1..1224
                         /organism="Gallus gallus"
                         /mol_type="genomic DNA"
    ...
    >>> name, content, _ = file[3]
    >>> print(name)
    VERSION
    >>> print(content)
    ['AJ311647.1  GI:13397825']
    >>> name, content, subfields = file[5]
    >>> print(name)
    SOURCE
    >>> print(content)
    ['Gallus gallus (chicken)']
    >>> print(dict(subfields))
    {'ORGANISM': ['Gallus gallus', 'Eukaryota; Metazoa; Chordata; ...', ...]}
    """

    def __init__(self):
        super().__init__()
        # Add '//' as general terminator of a GenBank file
        self.lines = ["//"]
        # Field start and stop indices in list of lines
        # and names of categories
        self._field_pos = []
        self._find_field_indices()
    
    @classmethod
    def read(cls, file):
        """
        Read a GenBank file.
        
        Parameters
        ----------
        file : file-like object or str
            The file to be read.
            Alternatively a file path can be supplied.
        
        Returns
        -------
        file_object : GenBankFile
            The parsed file.
        """
        file = super().read(file)
        file._find_field_indices()
        return file
    
    def get_fields(self, name):
        """
        Get all *GenBank* fields associated with a given field name.
        
        Parameters
        ----------
        name : str
            The field name.
        
        Returns
        -------
        fields : list of (list of str, OrderedDict of str -> str)
            A list containing the fields.
            For most field names, the list will only contain one
            element, but fields like *REFERENCE* are an exception.
            Each field is represented by a tuple.
            Each tuple contains as first element the content lines and
            as second element the subfields as dictionary.
            If the field has no subfields, the dictionary is empty.
        """
        indices = self.get_indices(name)
        # Omit the field name
        return [self[i][1:] for i in indices]
    
    def get_indices(self, name):
        """
        Get the indices to all *GenBank* fields associated with a given
        field name.
        
        Parameters
        ----------
        name : str
            The field name.
        
        Returns
        -------
        fields : list of int
            A list of indices.
            For most field names, the list will only contain one
            element, but fields like *REFERENCE* are an exception.
        """
        name = name.upper()
        indices = []
        for i, (_, _, fname) in enumerate(self._field_pos):
            if fname == name:
                indices.append(i)
        return indices
    
    def set_field(self, name, content, subfield_dict=None):
        """
        Set a *GenBank* field with the given content.

        If the field already exists in the file, the field is
        overwritten, otherwise a new field is created at the end of
        the file.
        
        Parameters
        ----------
        name : str
            The field name.
        content : list of str
            The content lines.
        subfield_dict : dict of str -> str, optional
            The subfields of the field.
            The dictionary maps subfield names to the content lines of
            the respective subfield.
        
        Raises
        ------
        InvalidFileError
            If the field occurs multiple times in the file.
            In this case it is ambiguous which field to overwrite.
        """
        name = name.upper()
        indices = self.get_indices(name)
        if len(indices) > 1:
            raise InvalidFileError(f"File contains multiple '{name}' fields")
        elif len(indices) == 1:
            # Replace existing entry
            index = indices[0]
            self[index] = name, content, subfield_dict
        else:
            # Add new entry as no entry exists yet
            self.append(name, content, subfield_dict)

    def __getitem__(self, index):
        index = self._translate_idx(index)
        start, stop, name = self._field_pos[index]
        
        if name in ["FEATURES", "ORIGIN"]:
            # For those two fields return the complete lines,
            # beginning with the line after the field name
            content = self._get_field_content(start+1, stop, indent=0)
            subfield_dict = OrderedDict()
        
        else:
            # For all metadata fields use the
            # standard GenBank indentation (=12)
            # Find subfields
            subfield_dict = OrderedDict()
            subfield_start = None
            first_subfield_start = None
            for i in range(start+1, stop):
                line = self.lines[i]
                # Check if line contains a new subfield
                # (Header beginning from first column)
                if len(line) != 0 and line[:12].strip() != "":
                    if first_subfield_start is None:
                        first_subfield_start = i
                    # Store previous subfield
                    if subfield_start is not None:
                        subfield_dict[header] = self._get_field_content(
                            subfield_start, i, indent=12
                        )
                    header = line[:12].strip()
                    subfield_start = i
            # Store last subfield
            if subfield_start is not None:
                subfield_dict[header] = self._get_field_content(
                    subfield_start, stop, indent=12
                )
            # Only include lines in field content,
            # that are not part of a subfield
            if first_subfield_start is not None:
                stop = first_subfield_start
            content = self._get_field_content(
                start, stop, indent=12
            )
        
        return name, content, subfield_dict
    
    def __setitem__(self, index, item):
        index = self._translate_idx(index)
        if not isinstance(item, tuple):
            raise TypeError(
                "Expected a tuple of name, content and optionally subfields"
            )
        if len(item) == 2:
            name, content = item
            subfields = None
        elif len(item) == 3:
            name, content, subfields = item
        else:
            raise TypeError(
                "Expected a tuple of name, content and optionally subfields"
            )
        inserted_lines = self._to_lines(name, content, subfields)
        
        # Stop of field to be replaced is start of new field
        start, old_stop, _ = self._field_pos[index]
        # If not the last element is set,
        # the following lines need to be added, too
        if old_stop is not len(self.lines):
            follow_lines = self.lines[old_stop:]
        else:
            follow_lines = []
        self.lines = self.lines[:start] + inserted_lines + follow_lines
        # Shift the start/stop indices of the following fields
        # by the amount of created fields
        shift = len(inserted_lines) - (old_stop - start)
        for i in range(index+1, len(self._field_pos)):
            old_start, old_stop, fname = self._field_pos[i]
            self._field_pos[i] = old_start+shift, old_stop+shift, fname
        # Add new entry
        self._field_pos[index] = start, start+len(inserted_lines), name.upper()
    
    def __delitem__(self, index):
        index = self._translate_idx(index)
        start, stop, _ = self._field_pos[index]
        # Shift the start/stop indices of the following fields
        # by the amount of deleted fields
        shift = stop - start
        for i in range(index, len(self._field_pos)):
            old_start, old_stop, name = self._field_pos[i]
            self._field_pos[i] = old_start-shift, old_stop-shift, name
        del self.lines[start : stop]
        del self._field_pos[index]
    
    def __len__(self):
        return len(self._field_pos)

    def insert(self, index, name, content, subfields=None):
        """
        Insert a *GenBank* field at the given position.
        
        Parameters
        ----------
        index : int
            The new field is inserted before the current field at this
            index.
            If the index is after the last field, the new field
            is appended to the end of the file.
        name : str
            The field name.
        content : list of str
            The content lines.
        subfield_dict : dict of str -> str, optional
            The subfields of the field.
            The dictionary maps subfield names to the content lines of
            the respective subfield.
        """
        index = self._translate_idx(index, length_exclusive=False)
        inserted_lines = self._to_lines(name, content, subfields)
        
        # Stop of previous field is start of new field
        if index == 0:
            start = 0
        else:
            _, start, _ = self._field_pos[index-1]
        # If the new lines are not inserted at the end,
        # the following lines need to be added, too
        if start is not len(self.lines):
            follow_lines = self.lines[start:]
        else:
            follow_lines = []
        self.lines = self.lines[:start] + inserted_lines + follow_lines
        # Shift the start/stop indices of the following fields
        # by the amount of created fields
        shift = len(inserted_lines)
        for i in range(index, len(self._field_pos)):
            old_start, old_stop, fname = self._field_pos[i]
            self._field_pos[i] = old_start+shift, old_stop+shift, fname
        # Add new entry
        self._field_pos.insert(
            index,
            (start, start+len(inserted_lines), name.upper())
        )
    
    def append(self, name, content, subfields=None):
        """
        Create a new *GenBank* field at the end of the file.
        
        Parameters
        ----------
        name : str
            The field name.
        content : list of str
            The content lines.
        subfield_dict : dict of str -> str, optional
            The subfields of the field.
            The dictionary maps subfield names to the content lines of
            the respective subfield.
        """
        self.insert(len(self), name, content, subfields)

    
    def _find_field_indices(self):
        """
        Identify the start and exclusive stop indices of lines
        corresponding to a field name for all fields in the file.
        """
        start = None
        name = ""
        self._field_pos = []
        for i, line in enumerate(self.lines):
            # Check if line contains a new major field
            # (Header beginning from first column)
            if len(line) != 0 and line[0] != " ":
                if line[:2] != "//":
                    stop = i
                    if start is not None:
                        # Store previous field
                        self._field_pos.append((start, stop, name))
                    start = i
                    name = line[0:12].strip()
                else:
                    # '//' means end of file
                    # -> Store last field
                    if start is not None:
                        stop = i
                        self._field_pos.append((start, stop, name))

    def _get_field_content(self, start, stop, indent):
        if indent == 0:
            return self.lines[start : stop]
        else:
            return [line[12:] for line in self.lines[start : stop]]
    
    def _to_lines(self, name, content, subfields):
        """
        Convert the field name, field content und subfield dictionary
        into text lines
        """
        if subfields is None:
            subfields = {}
        
        name = name.strip().upper()
        if len(name) == 0:
            raise ValueError(f"Must give a non emtpy name")
        subfields = OrderedDict({
            subfield_name.upper().strip() : subfield_lines
            for subfield_name, subfield_lines in subfields.items()
        })
        
        # Create lines for new field
        if name == "FEATURES":
            # Header line plus all actual feature lines
            lines = copy.copy(content)
            lines.insert(
                0, "FEATURES" + " "*13 + "Location/Qualifiers"
            )
        elif name == "ORIGIN":
            # Header line plus all actual sequence lines
            lines = copy.copy(content)
            lines.insert(0, "ORIGIN")
        else:
            name_column = []
            content_column = []
            # Create a line for the field name and empty lines
            # for each additional line required by the content 
            name_column += [name] + [""] * (len(content)-1)
            content_column += content
            for subfield_name, subfield_lines in subfields.items():
                name_column += ["  " + subfield_name] \
                               + [""] * (len(subfield_lines)-1)
                content_column += subfield_lines
            lines = [f"{n_col:12}{c_col}" for n_col, c_col
                              in zip(name_column, content_column)]
        
        return lines

    
    def _translate_idx(self, index, length_exclusive=True):
        """
        Check index boundaries and convert negative index to positive
        index.
        """
        if index < 0:
            new_index = len(self) + index
        else:
            new_index = index
        if length_exclusive:
            if new_index >= len(self):
                raise IndexError(f"Index {index} is out of range")
        else:
            if new_index > len(self):
                raise IndexError(f"Index {index} is out of range")
        return new_index


class MultiFile(TextFile):
    """
    This class represents a file in *GenBank* or *GenPept* format,
    that contains multiple entries, for more than one UID.
    
    The information for each UID are appended to each other in such a
    file.
    Objects of this class can be iterated to obtain a
    :class:`GenBankFile` for each entry in the file.
    
    Examples
    --------
    
    >>> import os.path
    >>> file_name = fetch_single_file(
    ...     ["1L2Y_A", "3O5R_A", "5UGO_A"],
    ...     os.path.join(path_to_directory, "multifile.gp"),
    ...     "protein", "gp"
    ... )
    >>> multi_file = MultiFile.read(file_name)
    >>> for gp_file in multi_file:
    ...     print(get_accession(gp_file))
    1L2Y_A
    3O5R_A
    5UGO_A
    """

    def __iter__(self):
        start_i = 0
        for i in range(len(self.lines)):
            line = self.lines[i]
            if line.strip() == "//":
                # Create file with lines corresponding to that file
                file_content = "\n".join(self.lines[start_i : i+1])
                file = GenBankFile.read(io.StringIO(file_content))
                # Reset file start index
                start_i = i
                yield file


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

from os.path import join
import glob
from subprocess import SubprocessError
import numpy as np
import pytest
import biotite.structure as struc
import biotite.structure.io as strucio
import biotite.structure.io.mmtf as mmtf
from biotite.application.dssp import DsspApp
from ..util import data_dir, is_not_installed


@pytest.mark.skipif(is_not_installed("mkdssp"), reason="DSSP is not installed")
@pytest.mark.parametrize(
    "path", glob.glob(join(data_dir("structure"), "*.mmtf"))
)
def test_dssp(path):
    sec_struct_codes = {0 : "I",
                        1 : "S",
                        2 : "H",
                        3 : "E",
                        4 : "G",
                        5 : "B",
                        6 : "T",
                        7 : "C"}

    mmtf_file = mmtf.MMTFFile.read(path)
    array = mmtf.get_structure(mmtf_file, model=1)
    array = array[array.hetero == False]
    first_chain_id = array.chain_id[0]
    chain = array[array.chain_id == first_chain_id]

    n_residues = struc.get_residue_count(chain)
    # Secondary structure annotation in PDB use also DSSP
    # -> compare PDB and local DSSP
    sse = mmtf_file["secStructList"]
    sse = sse[:n_residues]
    if (sse == -1).any():
        # First chain is not a pure polypeptide chain
        # (presumably DNA/RNA)
        # -> DSSP not applicable -> return
        return
    sse = np.array([sec_struct_codes[code] for code in sse],
                    dtype="U1")
    
    chain = array[array.chain_id == first_chain_id]
    sse_from_app = DsspApp.annotate_sse(chain)
    np.set_printoptions(threshold=10000)
    # PDB uses different DSSP version -> slight differences possible
    # -> only 90% must be identical
    assert np.count_nonzero(sse_from_app == sse) / len(sse) > 0.9


@pytest.mark.skipif(is_not_installed("mkdssp"), reason="DSSP is not installed")
def test_invalid_structure():
    array = strucio.load_structure(join(data_dir("structure"), "5ugo.mmtf"))
    # Get DNA chain -> Invalid for DSSP
    chain = array[array.chain_id == "T"]
    with pytest.raises(SubprocessError):
        DsspApp.annotate_sse(chain)




# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.application"
__author__ = "Patrick Kunzmann"
__all__ = ["MSAApp"]

import abc
from tempfile import NamedTemporaryFile
from collections import OrderedDict
import numpy as np
from .localapp import LocalApp, cleanup_tempfile
from .application import AppState, requires_state
from ..sequence.seqtypes import NucleotideSequence, ProteinSequence
from ..sequence.io.fasta.file import FastaFile
from ..sequence.align.alignment import Alignment
from .util import map_sequence, map_matrix


class MSAApp(LocalApp, metaclass=abc.ABCMeta):
    """
    This is an abstract base class for multiple sequence alignment
    software.
    
    It handles conversion of :class:`Sequence` objects to FASTA input
    and FASTA output to an :class:`Alignment` object.
    Inheriting subclasses only need to incorporate the file path
    of these FASTA files into the program arguments.

    Furthermore, this class can handle custom substitution matrices,
    if the underlying program supports these.

    MSA software that supports alignment of protein sequences and custom
    substitution matrices, can be used to align exotic, normally
    unsupported sequence types:
    At first the exotic sequences are mapped into protein sequences and
    the custom substitution matrix is converted into a protein sequence
    substitution matrix.
    Then the protein sequences are aligned and finally the protein
    sequences are mapped back into the original sequence types.
    The mapping does not work, when the alphabet of the exotic
    sequences is larger than the amino acid alphabet.
    
    Internally this creates a :class:`Popen` instance, which handles
    the execution.
    
    Parameters
    ----------
    sequences : iterable object of Sequence
        The sequences to be aligned.
    bin_path : str, optional
        Path of the MSA software binary. By default, the default path
        will be used.
    matrix : SubstitutionMatrix, optional
        A custom substitution matrix.
    """
    
    def __init__(self, sequences, bin_path, matrix=None):
        super().__init__(bin_path)
        
        if len(sequences) < 2:
            raise ValueError("At least two sequences are required")
        # Check if all sequences share the same alphabet
        alphabet = sequences[0].get_alphabet()
        for seq in sequences:
            if seq.get_alphabet() != alphabet:
                raise ValueError("Alphabets of the sequences are not equal")
        # Check matrix symmetry
        if matrix is not None and not matrix.is_symmetric():
            raise ValueError(
                "A symmetric matrix is required for "
                "multiple sequence alignments"
            )

        
        # Check whether the program supports the alignment for the given
        # sequence type
        if ProteinSequence.alphabet.extends(alphabet) \
            and self.supports_protein():
                self._is_mapped = False
                self._seqtype = "protein"
                if matrix is not None:
                    if not self.supports_custom_protein_matrix():
                        raise TypeError(
                            "The software does not support custom "
                            "substitution matrices for protein sequences"
                        )
                    self._matrix = matrix
                else:
                    self._matrix = None

        elif NucleotideSequence.alphabet_amb.extends(alphabet) \
            and self.supports_nucleotide():
                self._is_mapped = False
                self._seqtype = "nucleotide"
                if matrix is not None:
                    if not self.supports_custom_nucleotide_matrix():
                        raise TypeError(
                            "The software does not support custom "
                            "substitution matrices for nucleotide sequences"
                        )
                    self._matrix = matrix
                else:
                    self._matrix = None

        else:
            # For all other sequence types, try to map the sequence into
            # a protein sequence
            if not self.supports_protein():
                # Alignment of a custom sequence type requires mapping
                # into a protein sequence
                raise TypeError(
                    f"The software cannot align sequences of type "
                    f"{type(sequences[0]).__name__}: "
                    f"No support for alignment of the mapped sequences"
                )
            if not self.supports_custom_protein_matrix():
                # Alignment of a custom sequence type requires a custom
                # substitution matrix
                raise TypeError(
                    f"The software cannot align sequences of type "
                    f"{type(sequences[0]).__name__}: "
                    f"No support for custom substitution matrices"
                )
            self._is_mapped = True
            self._sequences = sequences
            # Sequence masquerades as protein
            self._seqtype = "protein"
            self._mapped_sequences = [
                map_sequence(sequence) for sequence in sequences
            ]
            self._matrix = map_matrix(matrix)


        self._sequences = sequences
        self._in_file = NamedTemporaryFile(
            "w", suffix=".fa", delete=False
        )
        self._out_file = NamedTemporaryFile(
            "r", suffix=".fa", delete=False
        )
        self._matrix_file = NamedTemporaryFile(
            "w", suffix=".mat", delete=False
        )

    def run(self):
        sequences = self._sequences if not self._is_mapped \
                    else self._mapped_sequences
        sequences_file = FastaFile()
        for i, seq in enumerate(sequences):
            sequences_file[str(i)] = str(seq)
        sequences_file.write(self._in_file)
        self._in_file.flush()
        if self._matrix is not None:
            self._matrix_file.write(str(self._matrix))
            self._matrix_file.flush()
        super().run()
    
    def evaluate(self):
        super().evaluate()
        alignment_file = FastaFile.read(self._out_file)
        seq_dict = OrderedDict(alignment_file)
        # Get alignment
        out_seq_str = [None] * len(seq_dict)
        for i in range(len(self._sequences)):
            out_seq_str[i] = seq_dict[str(i)]
        trace = Alignment.trace_from_strings(out_seq_str)
        self._alignment = Alignment(self._sequences, trace, None)
        # Also obtain original order
        self._order = np.zeros(len(seq_dict), dtype=int)
        for i, seq_index in enumerate(seq_dict):
             self._order[i] = int(seq_index)
    
    def clean_up(self):
        super().clean_up()
        cleanup_tempfile(self._in_file)
        cleanup_tempfile(self._out_file)
        cleanup_tempfile(self._matrix_file)
    
    @requires_state(AppState.JOINED)
    def get_alignment(self):
        """
        Get the resulting multiple sequence alignment.
        
        Returns
        -------
        alignment : Alignment
            The global multiple sequence alignment.
        """
        return self._alignment
    
    @requires_state(AppState.JOINED)
    def get_alignment_order(self):
        """
        Get the order of the resulting multiple sequence alignment.

        Usually the order of sequences in the output file is
        different from the input file, e.g. the sequences are ordered
        according to the guide tree.
        After running an MSA software, the output sequence order of
        the alignment rearranged so that it is the same as the input
        order.
        This method returns the order of the sequences intended by the
        MSA software.
        
        Returns
        -------
        order : ndarray, dtype=int
            The sequence order intended by the MSA software.
        
        Examples
        --------
        Align sequences and restore the original order:

        app = ClustalOmegaApp(sequences)
        app.start()
        app.join()
        alignment = app.get_alignment()
        order = app.get_alignment_order()
        alignment = alignment[:, order]
        """
        return self._order
    
    def get_input_file_path(self):
        """
        Get input file path (FASTA format).
        
        PROTECTED: Do not call from outside.
        
        Returns
        -------
        path : str
            Path of input file.
        """
        return self._in_file.name
    
    def get_output_file_path(self):
        """
        Get output file path (FASTA format).
        
        PROTECTED: Do not call from outside.
        
        Returns
        -------
        path : str
            Path of output file.
        """
        return self._out_file.name
    
    def get_matrix_file_path(self):
        """
        Get file path for custom substitution matrix.
        
        PROTECTED: Do not call from outside.
        
        Returns
        -------
        path : str or None
            Path of substitution matrix.
            None if no matrix was given.
        """
        return self._matrix_file.name if self._matrix is not None else None
    
    def get_seqtype(self):
        """
        Get the type of aligned sequences.

        When a custom sequence type (neither nucleotide nor protein)
        is mapped onto a protein sequence, the return value is also
        ``'protein'``.
        
        PROTECTED: Do not call from outside.
        
        Returns
        -------
        seqtype : {'nucleotide', 'protein'}
            Type of sequences to be aligned.
        """
        return self._seqtype
    
    @staticmethod
    @abc.abstractmethod
    def supports_nucleotide():
        """
        Check whether this class supports nucleotide sequences for
        alignment.

        Returns
        -------
        support : bool
            True, if the class has support, false otherwise.
        
        PROTECTED: Override when inheriting.
        """
        pass
    
    @staticmethod
    @abc.abstractmethod
    def supports_protein():
        """
        Check whether this class supports nucleotide sequences for
        alignment.

        Returns
        -------
        support : bool
            True, if the class has support, false otherwise.
        
        PROTECTED: Override when inheriting.
        """
        pass
    
    @staticmethod
    @abc.abstractmethod
    def supports_custom_nucleotide_matrix():
        """
        Check whether this class supports custom substitution matrices
        for protein sequence alignment.

        Returns
        -------
        support : bool
            True, if the class has support, false otherwise.
        
        PROTECTED: Override when inheriting.
        """
        pass
    
    @staticmethod
    @abc.abstractmethod
    def supports_custom_protein_matrix():
        """
        Check whether this class supports custom substitution matrices
        for nucleotide sequence alignment.

        Returns
        -------
        support : bool
            True, if the class has support, false otherwise.
        
        PROTECTED: Override when inheriting.
        """
        pass
    
    @classmethod
    def align(cls, sequences, bin_path=None, matrix=None):
        """
        Perform a multiple sequence alignment.
        
        This is a convenience function, that wraps the :class:`MSAApp`
        execution.
        
        Parameters
        ----------
        sequences : iterable object of Sequence
            The sequences to be aligned
        bin_path : str, optional
            Path of the MSA software binary. By default, the default
            path will be used.
        matrix : SubstitutionMatrix, optional
            A custom substitution matrix.
        
        Returns
        -------
        alignment : Alignment
            The global multiple sequence alignment.
        """
        if bin_path is None:
            app = cls(sequences, matrix=matrix)
        else:
            app = cls(sequences, bin_path, matrix=matrix)
        app.start()
        app.join()
        return app.get_alignment()
